# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from textwrap import dedent

from calmjs.parse import es5
from calmjs.parse.asttypes import Node
from calmjs.parse.asttypes import Identifier
from calmjs.parse.asttypes import Catch
from calmjs.parse.ruletypes import Attr
from calmjs.parse.ruletypes import Resolve
from calmjs.parse.ruletypes import Space
from calmjs.parse.ruletypes import RequiredSpace
from calmjs.parse.ruletypes import OpenBlock
from calmjs.parse.ruletypes import CloseBlock
from calmjs.parse.ruletypes import EndStatement

from calmjs.parse.unparsers.base import Dispatcher
from calmjs.parse.unparsers.walker import walk
from calmjs.parse.unparsers.es5 import Unparser
from calmjs.parse.handlers.indentation import indent
from calmjs.parse.handlers.core import rule_handler_noop
from calmjs.parse.handlers.core import token_handler_str_default
from calmjs.parse.handlers.core import layout_handler_space_minimum
from calmjs.parse.handlers.core import layout_handler_openbrace
from calmjs.parse.handlers.core import layout_handler_closebrace
from calmjs.parse.handlers.core import layout_handler_semicolon
from calmjs.parse.handlers.core import minimum_rules
from calmjs.parse.handlers.core import default_rules

from calmjs.parse.handlers.obfuscation import Scope
from calmjs.parse.handlers.obfuscation import CatchScope
from calmjs.parse.handlers.obfuscation import Obfuscator
from calmjs.parse.handlers.obfuscation import NameGenerator
from calmjs.parse.handlers.obfuscation import obfuscate
from calmjs.parse.handlers.obfuscation import token_handler_unobfuscate

empty_set = set({})


class NameGeneratorTestCase(unittest.TestCase):

    def test_basic(self):
        ng = NameGenerator()
        self.assertEqual('a', next(ng))
        self.assertEqual('b', next(ng))

    def test_skip(self):
        ng = NameGenerator(['if'], 'if')
        self.assertEqual('i', next(ng))
        self.assertEqual('f', next(ng))
        self.assertEqual('ii', next(ng))
        # if is skipped
        self.assertEqual('fi', next(ng))
        self.assertEqual('ff', next(ng))
        self.assertEqual('iii', next(ng))

    def test_additional_skip(self):
        ng1 = NameGenerator(['if'], 'if')
        ng2 = ng1(['ii'])
        v = iter(ng2)
        self.assertEqual('i', next(v))
        self.assertEqual('f', next(v))
        # ii is skipped
        # if is skipped
        self.assertEqual('fi', next(v))


class ScopeTestCase(unittest.TestCase):

    def test_scope_creation(self):
        scope = Scope(None)
        node = Node()
        scope.nest(node)
        self.assertEqual(scope.children[0].node, node)
        self.assertEqual(scope.children[0].parent, scope)

    def test_scope_symbols(self):
        scope = Scope(None)
        scope.declare('foo')
        scope.reference('foo')
        scope.reference('bar')
        self.assertEqual({'bar'}, scope.global_symbols)
        self.assertEqual(empty_set, scope.global_symbols_in_children)

    def test_resolve(self):
        scope = Scope(None)
        scope.remapped_symbols['foo'] = 'f'
        self.assertEqual('f', scope.resolve('foo'))
        self.assertEqual('bar', scope.resolve('bar'))

    def test_resolve_with_parent(self):
        root = Scope(None)
        root.remapped_symbols['root'] = 'r'
        root.remapped_symbols['name'] = 'root'
        child = root.nest(None)
        child.remapped_symbols['child'] = 'c'
        child.remapped_symbols['name'] = 'child'
        grandchild = child.nest(None)
        grandchild.remapped_symbols['grandchild'] = 'c'
        grandchild.remapped_symbols['name'] = 'grandchild'

        self.assertEqual('r', root.resolve('root'))
        self.assertEqual('root', root.resolve('name'))

        self.assertEqual('r', child.resolve('root'))
        self.assertEqual('child', child.resolve('name'))

        self.assertEqual('r', grandchild.resolve('root'))
        self.assertEqual('grandchild', grandchild.resolve('name'))

    def test_children_references(self):
        root = Scope(None)
        child1 = root.nest(None)
        grandchild1_1 = child1.nest(None)
        root.nest(None)  # child2
        child3 = root.nest(None)
        child3.nest(None)  # grandchild3_1
        grandchild3_2 = child3.nest(None)
        greatgrandchild3_2_1 = grandchild3_2.nest(None)

        self.assertEqual(empty_set, root.global_symbols_in_children)

        # one of the greatgrandchild make use of 'foo' without declaring
        # it.
        greatgrandchild3_2_1.reference('foo')
        self.assertEqual({'foo'}, root.global_symbols_in_children)
        self.assertEqual({'foo'}, child3.global_symbols_in_children)
        self.assertEqual(empty_set, child1.global_symbols_in_children)

        # some other grandchild does a few values, too
        grandchild1_1.reference('foo')
        self.assertEqual({'foo'}, child1.global_symbols_in_children)
        self.assertEqual({'foo'}, root.global_symbols_in_children)
        grandchild1_1.reference('a')
        self.assertEqual({'foo', 'a'}, child1.global_symbols_in_children)
        self.assertEqual({'foo', 'a'}, root.global_symbols_in_children)
        self.assertEqual({'foo'}, child3.global_symbols_in_children)

        # however, enough is enough, grandchild3_2 declares 'foo'
        grandchild3_2.declare('foo')
        # resulting in child3 no longer seeing that as referenced.
        self.assertEqual(empty_set, child3.global_symbols_in_children)
        # root still not immune as some other child still reference it.
        self.assertEqual({'foo', 'a'}, root.global_symbols_in_children)

    def test_close(self):
        root = Scope(None)
        root.reference('window')
        child = root.nest(None)
        child.reference('window')
        child.close()
        root.close()

        self.assertEqual({'window': 2}, root.referenced_symbols)
        self.assertEqual({'window': 1}, child.referenced_symbols)

        with self.assertRaises(ValueError):
            root.close()

        self.assertEqual({'window': 2}, root.referenced_symbols)

    def test_close_all_check_references(self):
        # for ease of counting everything
        root = Scope(None)
        child1 = root.nest(None)
        grandchild1_1 = child1.nest(None)
        child2 = root.nest(None)
        child3 = root.nest(None)
        child3.nest(None)  # grandchild3_1
        grandchild3_2 = child3.nest(None)
        greatgrandchild3_2_1 = grandchild3_2.nest(None)
        grandchild3_2.nest(None)  # greatgrandchild3_2_2
        child4 = root.nest(None)
        child5 = root.nest(None)

        greatgrandchild3_2_1.reference('window')
        greatgrandchild3_2_1.reference('window')
        grandchild3_2.reference('window')
        child3.declare('window')
        child3.reference('window')  # reference follows declare

        greatgrandchild3_2_1.reference('document')
        grandchild3_2.reference('document')

        child1.reference('window')
        grandchild1_1.reference('window')

        child2.reference('window')
        child2.reference('document')
        child2.declare('document')
        child2.reference('document')

        child4.reference('console')

        self.assertEqual({}, root.referenced_symbols)
        self.assertEqual({}, child5.referenced_symbols)
        root.close_all()
        self.assertEqual({
            'window': 3,
            'document': 2,
            'console': 1,
        }, root.referenced_symbols)
        self.assertEqual({
            'document': 2,
            'window': 1,
        }, child2.referenced_symbols)
        self.assertEqual({
            'document': 1,
            'window': 2,
        }, greatgrandchild3_2_1.referenced_symbols)
        self.assertEqual({
            'document': 2,
            'window': 3,
        }, grandchild3_2.referenced_symbols)

    def test_build_remap_symbols_parent_handling(self):
        # only after the final close is called.
        root = Scope(None)
        child1 = root.nest(None)
        child2 = root.nest(None)
        child3 = root.nest(None)
        grandchild2_1 = child2.nest(None)
        grandchild2_2 = child2.nest(None)
        grandchild2_3 = child2.nest(None)
        grandchild3_2 = child3.nest(None)
        greatgrandchild3_2_1 = grandchild3_2.nest(None)

        child3.declare('foo')
        child3.reference('foo')
        child2.declare('bar')
        child2.reference('bar')
        child2.reference('bar')  # to ensure this has priority
        child2.declare('foo')
        child2.reference('foo')
        # child1 does not declare fun
        child1.reference('fun')

        # first grandchild has declared and use foo, touches no
        # parents
        grandchild2_1.declare('foo')
        grandchild2_1.reference('foo')

        # the other grandchild declares new shadow, but also reference
        # a parent
        grandchild2_1.declare('bar')
        grandchild2_2.reference('bar')
        grandchild2_2.declare('custom')
        grandchild2_2.reference('custom')

        grandchild2_3.reference('foo')

        greatgrandchild3_2_1.declare('baz')
        greatgrandchild3_2_1.reference('window')
        root.declare('window')

        root.close_all()
        ng = NameGenerator()

        root.build_remap_symbols(ng)
        # root was ignored.
        self.assertEqual('window', root.resolve('window'))
        # these are independent.
        self.assertEqual('a', child3.resolve('foo'))
        self.assertEqual('a', child2.resolve('bar'))
        self.assertEqual('b', child2.resolve('foo'))

        # it never referenced no other variable
        self.assertEqual('a', grandchild2_1.resolve('foo'))
        # it has referenced bar, which was declared in parent ('a')
        self.assertEqual(child2.resolve('bar'), grandchild2_2.resolve('bar'))
        # naturally, its first declared variable is now 'b', shadows
        # over the remapped 'foo' which it doesn't use.
        self.assertEqual('b', grandchild2_2.resolve('custom'))

        # it will just get the foo fromchild2
        self.assertEqual('b', grandchild2_3.resolve('foo'))

        # this one was not remapped
        self.assertEqual('window', greatgrandchild3_2_1.resolve('window'))
        # remap the root node
        root.build_remap_symbols(ng, children_only=False)
        # this one was not remapped
        self.assertEqual('a', greatgrandchild3_2_1.resolve('window'))

    def test_build_remap_symbols_children_handling(self):
        ng = NameGenerator()
        root = Scope(None)
        child = root.nest(None)
        grandchild = child.nest(None)
        greatgrandchild = grandchild.nest(None)

        root.declare('bar')
        root.reference('bar')
        greatgrandchild.reference('a')
        greatgrandchild.declare('baz')
        greatgrandchild.reference('baz')

        root.build_remap_symbols(ng, children_only=False)
        # 'a' was taken by greatgrandchild referencing that as an
        # implicit global, so root scope must not redeclare 'a'.
        self.assertEqual('b', root.resolve('bar'))
        # the greatgrandchild should also not turn 'baz' into 'a' for
        # the same reason.
        self.assertEqual('b', greatgrandchild.resolve('baz'))

    def test_build_remap_not_shadowed(self):
        ng = NameGenerator()
        d0 = Scope(None)
        d1 = d0.nest(None)
        d2 = d1.nest(None)
        d3 = d2.nest(None)
        d4 = d3.nest(None)
        d5 = d4.nest(None)

        d0.declare('root')
        d0.reference('root')
        d1.declare('d1')
        d1.reference('d1')
        d2.declare('d2')
        d2.reference('d2')
        d3.declare('d3')
        d3.reference('d3')
        d4.declare('d4')
        d4.reference('d4')
        d5.declare('d5')
        d5.reference('d5')
        # and the innermost descedent references root at d0.
        d5.reference('root')

        d0.build_remap_symbols(ng, children_only=False)
        self.assertEqual('a', d0.resolve('root'))
        self.assertEqual('a', d1.resolve('d1'))
        self.assertEqual('a', d2.resolve('d2'))
        self.assertEqual('a', d5.resolve('root'))
        self.assertEqual('b', d5.resolve('d5'))

    def test_catch_scope_basic(self):
        with self.assertRaises(TypeError) as e:
            CatchScope(None, None)
        self.assertEqual(
            e.exception.args[0], 'CatchScopes must have a Scope as a parent')

        parent = Scope(None)
        with self.assertRaises(AttributeError):
            # None isn't a node that can be resolved.
            CatchScope(None, parent)

        with self.assertRaises(AttributeError):
            # doesn't work with generic node either
            CatchScope(Node(), parent)

        catcher = CatchScope(Catch(Identifier('e'), []), parent)
        self.assertEqual(catcher.catch_symbol, 'e')
        catcher.close()

        with self.assertRaises(ValueError):
            catcher.close()

    def test_catch_scope_interactions(self):
        parent = Scope(None)
        parent.declare('local')
        parent.reference('local')
        catcher = parent.catchctx(Catch(Identifier('exc'), []))
        self.assertEqual(catcher.catch_symbol, 'exc')
        catcher.reference('exc')
        self.assertEqual(catcher.catch_symbol_usage, 1)

        catcher.declare('caught')
        catcher.reference('caught')
        catcher.reference('global')
        catcher.reference('local')
        self.assertEqual({'caught', 'local'}, parent.local_declared_symbols)
        self.assertEqual(
            {'caught': 1, 'global': 1, 'local': 2},
            parent.referenced_symbols)
        self.assertEqual(
            {'caught': 1, 'global': 1, 'local': 2, 'exc': 1},
            catcher.referenced_symbols)
        self.assertEqual(
            {'caught', 'local', 'exc'}, catcher.local_declared_symbols)

        catcher_child = catcher.funcdecl(None)
        catcher_child.reference('global')
        catcher_child.reference('local')
        catcher_child.reference('exc')
        self.assertEqual(
            {'global': 1, 'local': 1, 'exc': 1},
            catcher_child.leaked_referenced_symbols)

        # these should remain unchanged for now.
        self.assertEqual(
            {'caught': 1, 'global': 1, 'local': 2},
            parent.referenced_symbols)
        self.assertEqual(
            {'caught': 1, 'global': 1, 'local': 2, 'exc': 1},
            catcher.referenced_symbols)

        parent.close_all()
        # parent should inherit everything
        self.assertEqual(
            {'caught': 1, 'global': 2, 'local': 3},
            parent.referenced_symbols)
        self.assertEqual(
            {'caught': 1, 'global': 2, 'local': 3, 'exc': 2},
            catcher.referenced_symbols)

        # build the table and try some lookup
        parent.build_remap_symbols(NameGenerator, False)
        # should not resolve into anything, 'exc' is exclusive to the
        # catch context and not its parent.
        self.assertEqual('exc', parent.resolve('exc'))
        # local has been referenced the most, and generated first
        self.assertEqual('a', parent.resolve('local'))
        # parent should be able to resolve the 'caught' variable as it
        # was declared in the catch context which is mirrored onto its
        # actual context.
        self.assertEqual('b', parent.resolve('caught'))
        # should resolve into c, since it cannot shadow the remapped
        # global that is used by its children, or as a matter of fact,
        # its actual scope.
        self.assertEqual('c', catcher.resolve('exc'))
        self.assertEqual('b', catcher.resolve('caught'))
        self.assertEqual('a', catcher.resolve('local'))
        # the child will be resolving the same values its parent, the
        # catch context, sees.
        self.assertEqual('c', catcher_child.resolve('exc'))
        self.assertEqual('b', catcher_child.resolve('caught'))
        self.assertEqual('a', catcher_child.resolve('local'))


class ObfuscatorTestCase(unittest.TestCase):

    maxDiff = None

    def test_simple_manual(self):
        tree = es5(dedent("""
        (function() {
          var foo = 1;
          var bar = 2;
          bar = 3;
        })(this);
        """).strip())
        obfuscator_unparser = Unparser(rules=(
            minimum_rules,
            obfuscate(),
        ))

        self.assertEqual(
            '(function(){var b=1;var a=2;a=3;})(this);',
            ''.join(c.text for c in obfuscator_unparser(tree)),
        )

    def test_multiple_reuse(self):
        tree = es5(dedent("""
        (function() {
          var foo = 1;
          var bar = 2;
          bar = 3;
        })(this);
        """).strip())
        obfuscator_unparser = Unparser(rules=(
            minimum_rules,
            obfuscate(),
        ))

        self.assertEqual(
            ''.join(c.text for c in obfuscator_unparser(tree)),
            ''.join(c.text for c in obfuscator_unparser(tree)),
        )

    def test_no_resolve(self):
        # a simple test to show that an obfuscator without the initial
        # loading run executed (i.e. the one with the required handlers)
        # with the resolve added to the dispatcher will not crash, but
        # simply not have any effect.

        tree = es5(dedent("""
        (function() {
          var foo = 1;
          var bar = 2;
        })(this);
        """).strip())
        unparser = Unparser()
        obfuscator = Obfuscator()
        dispatcher = Dispatcher(
            unparser.definitions, token_handler_str_default, {
                Space: layout_handler_space_minimum,
                RequiredSpace: layout_handler_space_minimum,
                OpenBlock: layout_handler_openbrace,
                CloseBlock: layout_handler_closebrace,
                EndStatement: layout_handler_semicolon,
            }, {
                Resolve: obfuscator.resolve,
            }
        )
        # see that the manually constructed minimum output works.
        self.assertEqual(
            "(function(){var foo=1;var bar=2;})(this);",
            ''.join(c.text for c in walk(dispatcher, tree))
        )

    def test_token_handler_unobfuscate(self):
        # simple test to test the handler works
        node = Identifier('dummy')
        token = Attr('value', pos=None)
        # when nothing got remapped.
        self.assertEqual(('dummy', None, None, None, None), next(
            token_handler_unobfuscate(token, None, node, 'dummy')))
        # when the provided value differs
        self.assertEqual(('d', None, None, 'dummy', None), next(
            token_handler_unobfuscate(token, None, node, 'd')))

    def test_build_substitutions(self):
        tree = es5(dedent("""
        (function(root) {
          var foo = 1;
          var bar = 2;
          baz = 3;
          foo = 4;
          window.document.body.focus();
        })(this, factory);
        """).strip())
        unparser = Unparser()
        obfuscator = Obfuscator()
        # a bare dispatcher should work, as it is used for extracting
        # the definitions from.
        sub_dispatcher = Dispatcher(
            unparser.definitions, rule_handler_noop, {}, {})
        # only do the intial walk.
        result = obfuscator.walk(sub_dispatcher, tree)
        # should be empty list as the run should produce nothing, due to
        # the null token producer.
        self.assertEqual(result, [])

        # only one scope was defined.
        self.assertEqual(1, len(obfuscator.scopes))
        self.assertEqual(1, len(obfuscator.global_scope.children))

        # do some validation on the scope itself.
        self.assertEqual(set(), obfuscator.global_scope.declared_symbols)
        self.assertEqual(
            {'factory': 1, 'baz': 1, 'window': 1},
            obfuscator.global_scope.referenced_symbols)
        scope = obfuscator.global_scope.children[0]

        self.assertEqual({'root', 'foo', 'bar'}, scope.declared_symbols)
        self.assertEqual({
            'root': 1,
            'foo': 2,
            'bar': 1,
            'baz': 1,
            'window': 1,
        }, scope.referenced_symbols)

        # do a trial run to show that the resolution works.
        main_dispatcher = Dispatcher(
            unparser.definitions, token_handler_unobfuscate, {
                Space: layout_handler_space_minimum,
                RequiredSpace: layout_handler_space_minimum,
                OpenBlock: layout_handler_openbrace,
                CloseBlock: layout_handler_closebrace,
                EndStatement: layout_handler_semicolon,
            }, {
                Resolve: obfuscator.resolve,
            }
        )
        # see that the manually constructed minimum output works.
        self.assertEqual(
            "(function(root){var foo=1;var bar=2;baz=3;foo=4;"
            "window.document.body.focus();})(this,factory);",
            ''.join(c.text for c in walk(main_dispatcher, tree))
        )
        # since nothing was remapped.
        self.assertEqual([
        ], [
            (c.text, c.name) for c in walk(main_dispatcher, tree)
            if c.name
        ])

        # now manually give the scope with a set of replacement names
        scope.remapped_symbols.update({
            'root': 'r',
            'foo': 'f',
            'bar': 'b',
            'baz': 'z',
        })
        self.assertEqual(
            "(function(r){var f=1;var b=2;z=3;f=4;"
            "window.document.body.focus();})(this,factory);",
            ''.join(c.text for c in walk(main_dispatcher, tree))
        )
        # check that the source chunks yielded contain both the original
        # text token and the original row/col location
        self.assertEqual([
            ('r', 1, 11, 'root'),
            ('f', 2, 7, 'foo'),
            ('b', 3, 7, 'bar'),
            ('z', 4, 3, 'baz'),
            ('f', 5, 3, 'foo'),
        ], [c[:4] for c in walk(main_dispatcher, tree) if c.name])

    def test_obfuscate_globals(self):
        node = es5(dedent("""
        var a_global = 1;
        (function(a_param) {
          var a_local = 1;
          a_local = a_param;
          a_local = a_global;
        })();
        """).strip())

        self.assertEqual(dedent("""
        var a_global = 1;
        (function(b) {
            var a = 1;
            a = b;
            a = a_global;
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(obfuscate_globals=False),
        ))(node)))

        self.assertEqual(dedent("""
        var a = 1;
        (function(c) {
            var b = 1;
            b = c;
            b = a;
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(obfuscate_globals=True),
        ))(node)))

    def test_obfuscate_no_global_recursive(self):
        node = es5(dedent("""
        (function named(param1, param2) {
          param1 = param1 * param2 - param2;
          param2--;
          if (param2 < 0) {
            return named(param1, param2);
          }
          return param1;
        })();
        """).strip())

        self.assertEqual(dedent("""
        (function named(b, a) {
          b = b * a - a;
          a--;
          if (a < 0) {
            return named(b, a);
          }
          return b;
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(obfuscate_globals=False),
        ))(node)))

    def test_obfuscate_no_global_recursive_redeclared_shadow_funcname(self):
        node = es5(dedent("""
        (function $() {
          $();
          (function $() {
            var foo = 1;
          })();
        })();
        """).strip())

        self.assertEqual(dedent("""
        (function $() {
          a();
          (function a() {
            var a = 1;
          })();
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(obfuscate_globals=False, shadow_funcname=True),
        ))(node)))

    def test_obfuscate_no_global_recursive_redeclared_no_shadow_funcname(self):
        node = es5(dedent("""
        (function $() {
          $();
          (function $() {
            var foo = 1;
          })();
        })();
        """).strip())

        self.assertEqual(dedent("""
        (function $() {
          a();
          (function a() {
            var b = 1;
          })();
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(obfuscate_globals=False, shadow_funcname=False),
        ))(node)))

    def test_obfuscate_shadow_funcname_not_mapped(self):
        node = es5(dedent("""
        function a(arg) {
        }
        """).strip())

        self.assertEqual(dedent("""
        function a(a) {
        }
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(obfuscate_globals=False, shadow_funcname=True),
        ))(node)))

    def test_obfuscate_no_shadow_funcname_not_mapped(self):
        node = es5(dedent("""
        function a(arg) {
        }
        """).strip())

        self.assertEqual(dedent("""
        function a(b) {
        }
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(obfuscate_globals=False, shadow_funcname=False),
        ))(node)))

    def test_obfuscate_skip(self):
        node = es5(dedent("""
        (function(a_param) {
          var a_local = a_param;
        })();
        """).strip())

        self.assertEqual(dedent("""
        (function(c) {
            var d = c;
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(reserved_keywords=('a', 'b',)),
        ))(node)))

    def test_no_shadow_children(self):
        node = es5(dedent("""
        var some_value = 1;
        (function(param) {
          a.is_not_declared_in_this_file(param);
        })();
        """).strip())

        self.assertEqual(dedent("""
        var b = 1;
        (function(b) {
            a.is_not_declared_in_this_file(b);
        })();
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(obfuscate_globals=True),
        ))(node)))

    def test_no_shadow_parent_remapped(self):
        node = es5(dedent("""
        var a = c;
        (function(param) {
          b.is_not_declared_in_this_file(a, param);
        })(c);
        """).strip())

        # param can remap to c because the global 'c' not used in that
        # scope.
        self.assertEqual(dedent("""
        var a = c;
        (function(c) {
            b.is_not_declared_in_this_file(a, c);
        })(c);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(obfuscate_globals=True),
        ))(node)))

    def test_multi_scope(self):
        node = es5(dedent("""
        (function(value) {
          var parent = 1;
          (function(value) {
            var child = 2;
            (function(value) {
              var grandchild = 3;
              (function(value) {
                var greatgrandchild = 4;
                (function(value) {
                  var result = 1;
                  // using greatgrandchild a lot to ensure priority
                  // mucking
                  console.log(parent);
                  console.log(greatgrandchild);
                  console.log(greatgrandchild * value);
                  console.log(greatgrandchild * parent);
                  console.log(greatgrandchild * greatgrandchild);
                })(value);
              })(value);
            })(value);
          })(value);
        })(0);
        """).strip())

        # Note that grandchild and greatgrandchild never had any of
        # their local variables referenced by any of their nested
        # scopes, their values got shadowed in the greatgrandchild scope
        # before both that and parent is used in the innermost scope.
        # Also, note that since greatgrandchild stopped propagating the
        # reference usage upwards, even though it has a lot more usage,
        # it never will claim priority over the parent at the parent
        # scope since parent got mapped to 'a' thus forcing
        # greatgrandchild to map to 'b', the next lowest value symbol.
        self.assertEqual(dedent("""
        (function(b) {
          var a = 1;
          (function(b) {
            var c = 2;
            (function(b) {
              var c = 3;
              (function(c) {
                var b = 4;
                (function(c) {
                  var d = 1;
                  console.log(a);
                  console.log(b);
                  console.log(b * c);
                  console.log(b * a);
                  console.log(b * b);
                })(c);
              })(b);
            })(b);
          })(b);
        })(0);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))

    def test_obfuscate_try_catch_shadowed(self):
        node = es5(dedent("""
        var value = 1;
        try {
          console.log(value);
          throw Error("welp");
        }
        catch (value) {
          console.log(value);
        }
        """).strip())

        self.assertEqual(dedent("""
        var a = 1;
        try {
            console.log(a);
            throw Error("welp");
        }
        catch (a) {
            console.log(a);
        }
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='    '),
            obfuscate(obfuscate_globals=True),
        ))(node)))

    def test_obfuscate_try_catch_shadowed_scoped(self):
        node = es5(dedent("""
        var value = 100;
        (function() {
          var value = 1;
          try {
            console.log(value);
            throw Error("welp");
          }
          catch (value) {
            console.log(value);
          }
        })();
        console.log(value);
        """).strip())

        self.assertEqual(dedent("""
        var value = 100;
        (function() {
          var a = 1;
          try {
            console.log(a);
            throw Error("welp");
          }
          catch (a) {
            console.log(a);
          }
        })();
        console.log(value);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))

    def test_obfuscate_try_catch_no_declare(self):
        node = es5(dedent("""
        var value = 100;
        (function() {
          value = 1;
          try {
            console.log(value);
            throw Error("welp");
          }
          catch (value) {
            value = 0;
            console.log(value);
          }
          console.log(value);
        })();
        console.log(value);
        """).strip())

        self.assertEqual(dedent("""
        var value = 100;
        (function() {
          value = 1;
          try {
            console.log(value);
            throw Error("welp");
          }
          catch (a) {
            a = 0;
            console.log(a);
          }
          console.log(value);
        })();
        console.log(value);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))

    def test_obfuscate_try_catch_vardeclare_inside_catch(self):
        node = es5(dedent("""
        var value = 100;
        (function() {
          console.log(value);
          value = 1;
          try {
            console.log(value);
            throw Error("welp");
          }
          catch (exc) {
            var value = 2;
            console.log(value);
          }
          value = 3;
        })();
        console.log(value);
        """).strip())

        self.assertEqual(dedent("""
        var value = 100;
        (function() {
          console.log(a);
          a = 1;
          try {
            console.log(a);
            throw Error("welp");
          }
          catch (b) {
            var a = 2;
            console.log(a);
          }
          a = 3;
        })();
        console.log(value);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))

    def test_obfuscate_try_catch_scope_inside_catch(self):
        node = es5(dedent("""
        var value = 100;
        (function() {
          var dummy = 0;
          console.log(value);
          value = 1;
          try {
            console.log(value);
            throw Error("welp");
          }
          catch (exc) {
            var value = 2;
            var welped = value;
            welped = 'welped';
            (function() {
              console.log(exc);
              console.log(value);
              console.log(welped);
            })();
          }
          value = 3;
        })();
        console.log(value);
        """).strip())

        # note that dummy -> c because exc is a local, not counted in
        # the priority at the parent scope.
        self.assertEqual(dedent("""
        var value = 100;
        (function() {
          var c = 0;
          console.log(a);
          a = 1;
          try {
            console.log(a);
            throw Error("welp");
          }
          catch (d) {
            var a = 2;
            var b = a;
            b = 'welped';
            (function() {
              console.log(d);
              console.log(a);
              console.log(b);
            })();
          }
          a = 3;
        })();
        console.log(value);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))

    def test_functions_in_lists(self):
        node = es5(dedent("""
        (function main(root) {
          root.exports = [
            (function(module, exports) {
              module.exports = {};
            }),
            (function(module, exports) {
              exports.fun = 1;
            }),
          ];
        })(this);
        """).strip())

        self.assertEqual(dedent("""
        (function main(a) {
          a.exports = [(function(a, b) {
            a.exports = {};
          }), (function(b, a) {
            a.fun = 1;
          })];
        })(this);
        """).lstrip(), ''.join(c.text for c in Unparser(rules=(
            default_rules,
            indent(indent_str='  '),
            obfuscate(),
        ))(node)))



# -*- coding: utf-8 -*-

import unittest
import importlib
import os
import sys

from io import StringIO
from shutil import rmtree
from tempfile import mkdtemp
from types import ModuleType
from ply import lex
from calmjs.parse.parsers import optimize
from calmjs.parse.parsers import es5
from calmjs.parse.utils import ply_dist


class OptimizeTestCase(unittest.TestCase):

    def setUp(self):
        self.purged = []
        optimize.unlink = self.purged.append

    def tearDown(self):
        optimize.unlink = os.unlink
        optimize.import_module = importlib.import_module
        optimize.ply_dist = ply_dist
        # undo whatever monkey patch that may have happened
        lex.open = open

    def break_ply(self):
        import ply

        def cleanup():
            sys.modules['ply'] = ply

        self.addCleanup(cleanup)
        sys.modules['ply'] = None

    def test_verify_paths(self):
        tempdir = mkdtemp()
        self.addCleanup(rmtree, tempdir)

        # create fake module files
        modules = [os.path.join(tempdir, name) for name in (
            'foo.pyc', 'bar.py', 'bar.pyc', 'foo.py')]

        for module in modules:
            with open(module, 'w'):
                pass

        self.assertEqual(
            sorted(modules),
            sorted(optimize.verify_paths(modules[:2])),
        )

    def test_find_tab_paths(self):
        fake_es5 = ModuleType('fake_es5')
        fake_es5.lextab = 'some_lextab'
        fake_es5.yacctab = 'some_yacctab'
        paths, missing = optimize.find_tab_paths(fake_es5)
        self.assertEqual([], paths)
        self.assertEqual(['some_lextab', 'some_yacctab'], missing)

        # ensure the parser exists
        es5.Parser()
        # should have created the optimized version of the file, if not
        # already exists
        answers = [
            sys.modules[es5.lextab].__file__,
            sys.modules[es5.yacctab].__file__,
        ]
        paths, missing = optimize.find_tab_paths(es5)
        self.assertEqual(paths, answers)
        self.assertEqual([], missing)

    def test_unlink_modules(self):
        es5.Parser()
        p = sys.modules[es5.yacctab].__file__
        self.assertTrue(os.path.exists(p))
        # unlink has been patched out
        optimize.purge_tabs(es5)
        self.assertNotIn(es5.yacctab, sys.modules)
        self.assertNotEqual(len(self.purged), 0)

    def test_not_imported(self):
        # mock the import failures
        def fail_import(module):
            raise ImportError()
        optimize.import_module = fail_import
        optimize.purge_tabs(es5)
        # since none are imported, no purges.
        self.assertEqual(len(self.purged), 0)

    def test_safeguard(self):
        fake_es5 = ModuleType('fake_es5')
        fake_es5.lextab = 'some_lextab'
        fake_es5.yacctab = 'some.system.module'

        with self.assertRaises(ValueError):
            optimize.purge_tabs(fake_es5)
        # since none are imported, no purges.
        self.assertEqual(len(self.purged), 0)

    def test_reoptimize(self):
        # this ensures everything is created
        optimize.reoptimize_all()
        # reset the purged list
        self.purged[:] = []
        # do the "real" run, should be twice the number of implemented
        # parsers.
        optimize.reoptimize_all()
        self.assertNotEqual(len(self.purged), 0)
        self.assertIs(lex.open, open)

    def test_reoptimize_monkey_patched(self):
        optimize.reoptimize_all(True)
        self.assertIsNot(lex.open, open)
        self.assertNotEqual(len(self.purged), 0)

    def test_optimize_build(self):
        called = []

        def sentinel(*a, **kw):
            called.append(True)

        fake_es5 = ModuleType('fake_namespace.fake_es5')
        fake_es5.Parser = sentinel

        # inject fake namespace and module
        sys.modules['fake_namespace'] = ModuleType('fake_namespace')
        self.addCleanup(sys.modules.pop, 'fake_namespace')
        sys.modules['fake_namespace.fake_es5'] = fake_es5
        self.addCleanup(sys.modules.pop, 'fake_namespace.fake_es5')

        optimize.optimize_build('fake_namespace.fake_es5')
        self.assertEqual(len(self.purged), 0)
        self.assertTrue(called)

    def test_optimize_first_build(self):
        optimize.reoptimize_all(True, first_build=True)
        # shouldn't have purged any modules
        self.assertEqual(len(self.purged), 0)

    def test_optimize_first_build_valid_with_broken_ply(self):
        self.break_ply()
        optimize.reoptimize_all(True, first_build=True)
        # shouldn't have purged any modules
        self.assertEqual(len(self.purged), 0)

    def test_assume_ply_version_default_ply(self):
        # only applicable if no ply_dist found
        optimize.ply_dist = None
        stderr = sys.stderr
        self.addCleanup(setattr, sys, 'stderr', stderr)

        # where ply is actually available; and since the real thing is
        # expected to be present and usable, intersperse that real value
        # into the expected string.
        import ply
        sys.stderr = StringIO()
        optimize._assume_ply_version()
        self.assertTrue(sys.stderr.getvalue().startswith(
            "WARNING: cannot find distribution for 'ply'; using value "
            "provided by ply, assuming 'ply==%s' for pre-generated modules" % (
                ply.__version__
            )))

    def test_assume_ply_version_override_ply(self):
        # can still override if ply is actually available
        optimize.ply_dist = None
        stderr = sys.stderr
        self.addCleanup(setattr, sys, 'stderr', stderr)

        self.addCleanup(os.environ.pop, optimize._ASSUME_ENVVAR, None)
        sys.stderr = StringIO()
        os.environ[optimize._ASSUME_ENVVAR] = '0.9999'  # should never exist
        optimize._assume_ply_version()
        self.assertTrue(sys.stderr.getvalue().startswith(
            "WARNING: cannot find distribution for 'ply'; using environment "
            "variable 'CALMJS_PARSE_ASSUME_PLY_VERSION', "
            "assuming 'ply==0.9999' for pre-generated modules"))

    def test_assume_ply_version_no_ply(self):
        # default when ply is fully broken.
        optimize.ply_dist = None
        stderr = sys.stderr
        self.addCleanup(setattr, sys, 'stderr', stderr)

        self.break_ply()
        sys.stderr = StringIO()
        optimize._assume_ply_version()
        self.assertTrue(sys.stderr.getvalue().startswith(
            "WARNING: cannot find distribution for 'ply'; using default "
            "value, assuming 'ply==3.11' for pre-generated modules"))

    def test_optimize_first_build_valid_with_broken_ply_error(self):
        def fail_import(*a, **kw):
            raise ImportError('no module named ply')

        optimize.import_module = fail_import

        with self.assertRaises(ImportError):
            optimize.reoptimize_all()

        stderr = sys.stderr

        def cleanup():
            sys.stderr = stderr

        self.addCleanup(cleanup)

        sys.stderr = StringIO()
        with self.assertRaises(SystemExit):
            optimize.reoptimize_all(first_build=True)

        self.assertTrue(sys.stderr.getvalue().startswith(
            "ERROR: cannot find pre-generated modules for the assumed 'ply' "
            "version"))

    def test_optimize_first_build_assume_broken_ply_error(self):
        optimize.ply_dist = None

        self.break_ply()

        def fail_import(*a, **kw):
            raise ImportError('no module named ply')

        optimize.import_module = fail_import

        with self.assertRaises(ImportError):
            optimize.reoptimize_all()

        stderr = sys.stderr

        def cleanup():
            sys.stderr = stderr

        self.addCleanup(cleanup)

        sys.stderr = StringIO()
        with self.assertRaises(SystemExit):
            optimize.reoptimize_all(first_build=True)

        lines = sys.stderr.getvalue().splitlines()
        self.assertTrue(lines[0].startswith(
            "WARNING: cannot find distribution for 'ply'; using default value"
            ))
        self.assertTrue(lines[1].startswith(
            "ERROR: cannot find pre-generated modules for the assumed 'ply' "
            "version"))

    def test_optimize_build_assume_broken_ply_but_available(self):
        optimize.ply_dist = None
        called = []

        def sentinel(*a, **kw):
            called.append(True)

        fake_es5 = ModuleType('fake_namespace.fake_es5')
        fake_es5.Parser = sentinel

        # inject fake namespace and module
        sys.modules['fake_namespace'] = ModuleType('fake_namespace')
        self.addCleanup(sys.modules.pop, 'fake_namespace')
        sys.modules['fake_namespace.fake_es5'] = fake_es5
        self.addCleanup(sys.modules.pop, 'fake_namespace.fake_es5')
        stderr = sys.stderr
        self.addCleanup(setattr, sys, 'stderr', stderr)
        sys.stderr = StringIO()

        optimize.optimize_build('fake_namespace.fake_es5')
        self.assertEqual(len(self.purged), 0)
        self.assertTrue(called)
        # this parser will not actually error as it does nothing; and
        # so not actually care whether ply actually available here or
        # not.
        self.assertTrue(sys.stderr.getvalue().startswith(
            "WARNING: cannot find distribution for 'ply'; "
            ))
        self.assertNotIn('ERROR', sys.stderr.getvalue())



# -*- coding: utf-8 -*-
"""
Source map helpers
"""

from __future__ import unicode_literals, absolute_import
import base64
import json
import logging
from os.path import sep

from calmjs.parse.vlq import encode_mappings
from calmjs.parse.utils import normrelpath

logger = logging.getLogger(__name__)

# for NotImplemented source values
INVALID_SOURCE = 'about:invalid'
default_encoding = 'utf8'


class Names(object):
    """
    A class for tracking and reporting of names for usage with source
    maps.
    """

    def __init__(self):
        self._names = {}
        self._current = 0

    def update(self, name):
        """
        Query a name for the relative index value to be added into the
        source map name field (optional 5th element).
        """

        if name is None:
            return

        if name not in self._names:
            # add the name if it isn't already tracked
            self._names[name] = len(self._names)

        result = self._names[name] - self._current
        self._current = self._names[name]
        return result

    def __iter__(self):
        for name, idx in sorted(self._names.items(), key=lambda x: x[1]):
            yield name


class Bookkeeper(object):
    """
    A class for tracking positions

    Set a current position, read out a delta compared to the previous
    for a given attribute
    """

    def __init__(self):
        super(Bookkeeper, self).__setattr__('_prev', {})
        super(Bookkeeper, self).__setattr__('_curr', {})

    def _hasattr(self, attr):
        return all(
            isinstance(check.get(attr, None), int)
            for check in (self._prev, self._curr)
        )

    def __setattr__(self, attr, value):
        """
        Set the current position
        """

        chk = attr[:1] == '_'
        attr = attr[1:] if chk else attr

        if not isinstance(value, int):
            raise TypeError("assignment must be of type 'int'")

        if not self._hasattr(attr) or chk:
            self._curr[attr] = self._prev[attr] = value
        else:
            self._curr[attr], self._prev[attr] = value, self._curr[attr]

    def __getattr__(self, attr):
        chk = attr[:1] == '_'
        attr = attr[1:] if chk else attr
        if not self._hasattr(attr):
            raise AttributeError("'%s' object has no attribute %r" % (
                self.__class__.__name__, attr))
        return self._curr[attr] if chk else self._curr[attr] - self._prev[attr]

    def __delattr__(self, attr):
        if not self._hasattr(attr):
            raise AttributeError("'%s' object has no attribute %r" % (
                self.__class__.__name__, attr))
        self._prev[attr] = self._curr[attr] = 0


class Book(object):
    """
    For storing calculated offsets, if required.
    """

    def __init__(self, bookkeeper):
        # length of previously written chunk.text.
        self.written_len = 0
        # length of original text for previously written chunk.text
        self.original_len = 0
        self.keeper = bookkeeper


def default_book():
    bk = Bookkeeper()
    # index of the current file can be implemented/tracked with the
    # Names class.

    # position of the current line that is being written; 0-indexed as
    # there are no existing requirements, and that it maps directly to
    # the length of the string written (usually).
    bk.sink_column = 0
    # since the source line/col positions have been implemented as
    # 1-indexed values, so the offset is pre-applied like so.
    bk.source_line = 1
    bk.source_column = 1
    return Book(bk)


def normalize_mapping_line(mapping_line, previous_source_column=0):
    """
    Often times the position will remain stable, such that the naive
    process will end up with many redundant values; this function will
    iterate through the line and remove all extra values.
    """

    if not mapping_line:
        return [], previous_source_column

    # Note that while the local record here is also done as a 4-tuple,
    # element 1 and 2 are never used since they are always provided by
    # the segments in the mapping line; they are defined for consistency
    # reasons.

    def regenerate(segment):
        if len(segment) == 5:
            result = (record[0], segment[1], segment[2], record[3], segment[4])
        else:
            result = (record[0], segment[1], segment[2], record[3])
        # Ideally the exact location should still be kept, but given
        # that the sourcemap format is accumulative and permits a lot
        # of inferred positions, resetting all values to 0 is intended.
        record[:] = [0, 0, 0, 0]
        return result

    # first element of the line; sink column (0th element) is always
    # the absolute value, so always use the provided value sourced from
    # the original mapping_line; the source column (3rd element) is
    # never reset, so if a previous counter exists (which is specified
    # by the optional argument), make use of it to generate the initial
    # normalized segment.
    record = [0, 0, 0, previous_source_column]
    result = []
    regen_next = True

    for segment in mapping_line:
        if not segment:
            # ignore empty records
            continue
        # if the line has not changed, and that the increases of both
        # columns are the same, accumulate the column counter and drop
        # the segment.

        # accumulate the current record first
        record[0] += segment[0]
        if len(segment) == 1:
            # Mark the termination, as 1-tuple determines the end of the
            # previous symbol and denote that whatever follows are not
            # in any previous source files.  So if it isn't recorded,
            # make note of this if it wasn't done already.
            if result and len(result[-1]) != 1:
                result.append((record[0],))
                record[0] = 0
                # the next complete segment will require regeneration
                regen_next = True
            # skip the remaining processing.
            continue

        record[3] += segment[3]

        # 5-tuples are always special case with the remapped identifier
        # name element, and to mark the termination the next token must
        # also be explicitly written (in our case, regenerated).  If the
        # filename or source line relative position changed (idx 1 and
        # 2), regenerate it too.  Finally, if the column offsets differ
        # between source and sink, regenerate.
        if len(segment) == 5 or regen_next or segment[1] or segment[2] or (
                record[0] != record[3]):
            result.append(regenerate(segment))
            regen_next = len(segment) == 5

    # must return the consumed/omitted values.
    return result, record[3]


def normalize_mappings(mappings, column=0):
    result = []
    for ml in mappings:
        new_ml, column = normalize_mapping_line(ml, column)
        result.append(new_ml)
    return result


def write(
        stream_fragments, stream, normalize=True,
        book=None, sources=None, names=None, mappings=None):
    """
    Given an iterable of stream fragments, write it to the stream object
    by using its write method.  Returns a 3-tuple, where the first
    element is the mapping, second element is the list of sources and
    the third being the original names referenced by the given fragment.

    Arguments:

    stream_fragments
        an iterable that only contains StreamFragments
    stream
        an io.IOBase compatible stream object
    normalize
        the default True setting will result in the mappings that were
        returned be normalized to the minimum form.  This will reduce
        the size of the generated source map at the expense of slightly
        lower quality.

        Also, if any of the subsequent arguments are provided (for
        instance, for the multiple calls to this function), the usage of
        the normalize flag is currently NOT supported.

        If multiple sets of outputs are to be produced, the recommended
        method is to chain all the stream fragments together before
        passing in.

    Advanced usage arguments

    book
        A Book instance; if none is provided an instance will be created
        from the default_book constructor.  The Bookkeeper instance is
        used for tracking the positions of rows and columns of the input
        stream.
    sources
        a Names instance for tracking sources; if None is provided, an
        instance will be created for internal use.
    names
        a Names instance for tracking names; if None is provided, an
        instance will be created for internal use.
    mappings
        a previously produced mappings.

    A stream fragment tuple must contain the following

    - The string to write to the stream
    - Original starting line of the string; None if not present
    - Original starting column fo the line; None if not present
    - Original string that this fragment represents (i.e. for the case
      where this string fragment was an identifier but got mangled into
      an alternative form); use None if this was not the case.
    - The source of the fragment.  If the first fragment is unspecified,
      the INVALID_SOURCE url will be used (i.e. about:invalid).  After
      that, a None value will be treated as the implicit value, and if
      NotImplemented is encountered, the INVALID_SOURCE url will be used
      also.

    If a number of stream_fragments are to be provided, common instances
    of Book (constructed via default_book) and Names (for sources and
    names) should be provided if they are not chained together.
    """

    def push_line():
        mappings.append([])
        book.keeper._sink_column = 0

    if names is None:
        names = Names()

    if sources is None:
        sources = Names()

    if book is None:
        book = default_book()

    if not isinstance(mappings, list):
        # note that
        mappings = []
        # finalize initial states; the most recent list (mappings[-1])
        # is the current line
        push_line()

    for chunk, lineno, colno, original_name, source in stream_fragments:
        # note that lineno/colno are assumed to be both provided or none
        # provided.
        lines = chunk.splitlines(True)
        for line in lines:
            stream.write(line)

            # Two separate checks are done.  As per specification, if
            # either lineno or colno are unspecified, it is assumed that
            # the segment is unmapped - append a termination (1-tuple)
            #
            # Otherwise, note that if this segment is the beginning of a
            # line, and that an implied source colno/linecol were
            # provided (i.e. value of 0), and that the string is empty,
            # it can be safely skipped, since it is an implied and
            # unmapped indentation

            if lineno is None or colno is None:
                mappings[-1].append((book.keeper.sink_column,))
            else:
                name_id = names.update(original_name)
                # this is a bit of a trick: an unspecified value (None)
                # will simply be treated as the implied value, hence 0.
                # However, a NotImplemented will be recorded and be
                # convereted to the invalid url at the end.
                source_id = sources.update(source) or 0

                if lineno:
                    # a new lineno is provided, apply it to the book and
                    # use the result as the written value.
                    book.keeper.source_line = lineno
                    source_line = book.keeper.source_line
                else:
                    # no change in offset, do not calculate and assume
                    # the value to be written is unchanged.
                    source_line = 0

                # if the provided colno is to be inferred, calculate it
                # based on the previous line length plus the previous
                # real source column value, otherwise standard value
                # for tracking.

                # the reason for using the previous lengths is simply
                # due to how the bookkeeper class does the calculation
                # on-demand, and that the starting column for the
                # _current_ text fragment can only be calculated using
                # what was written previously, hence the original length
                # value being added if the current colno is to be
                # inferred.
                if colno:
                    book.keeper.source_column = colno
                else:
                    book.keeper.source_column = (
                        book.keeper._source_column + book.original_len)

                if original_name is not None:
                    mappings[-1].append((
                        book.keeper.sink_column, source_id,
                        source_line, book.keeper.source_column,
                        name_id
                    ))
                else:
                    mappings[-1].append((
                        book.keeper.sink_column, source_id,
                        source_line, book.keeper.source_column
                    ))

            # doing this last to update the position for the next line
            # or chunk for the relative values based on what was added
            if line[-1:] in '\r\n':
                colno = (
                    colno if colno in (0, None) else
                    colno + len(line.rstrip()))
                book.original_len = book.written_len = 0
                push_line()

                if lineno and colno:
                    # naturally, a provided lineno and colno can be
                    # safely inferred
                    lineno += 1
                    colno = 1
                    continue

                # This normally shouldn't happen with sane parsers
                # and lexers, but this assumes that no further symbols
                # aside from the new lines got inserted.  So this is
                # likely caused by some generated element produced
                # inferred fragments that include newlines, and without
                # the exact location the chunk cannot be manually
                # tracked.  Simply warn about this edge case and
                # continue processing.
                if line is not lines[-1]:
                    logger.warning(
                        'text in the generated stream at line %d may be '
                        'mapped incorrectly due to stream fragment containing '
                        'a trailing newline character provided without both '
                        'lineno and colno defined; '
                        'text fragment originated from: %s',
                        len(mappings),
                        source if source else '<unknown>',
                    )
                    logger.info(
                        'text in stream fragments should not have trailing '
                        'characters after a new line, they should be split '
                        'off into a separate fragment.'
                    )
            else:
                book.written_len = len(line)
                book.original_len = (
                    len(original_name) if original_name else book.written_len)
                book.keeper.sink_column = (
                    book.keeper._sink_column + book.written_len)

    # normalize everything
    if normalize:
        # if this _ever_ supports the multiple usage using existence
        # instances of names and book and mappings, it needs to deal
        # with NOT normalizing the existing mappings and somehow reuse
        # the previously stored value, probably in the book.  It is
        # most certainly a bad idea to support that use case while also
        # supporting the default normalize flag due to the complex
        # tracking of all the existing values...
        mappings = normalize_mappings(mappings)

    list_sources = [
        INVALID_SOURCE if s == NotImplemented else s for s in sources
    ] or [INVALID_SOURCE]
    return mappings, list_sources, list(names)


def encode_sourcemap(filename, mappings, sources, names=[]):
    """
    Take a filename, mappings and names produced from the write function
    and sources.  As the write function currently does not handle the
    tracking of source filenames, the sources should be a list of one
    element with the original filename.

    Arguments

    filename
        The target filename that the stream was or to be written to.
        The stream being the argument that was supplied to the write
        function
    mappings
        The raw unencoded mappings produced by write, which is returned
        as its second element.
    sources
        List of original source filenames.  When used in conjunction
        with the above write function, it should be a list of one item,
        being the path to the original filename.
    names
        The list of original names generated by write, which is returned
        as its first element.

    Returns a dict which can be JSON encoded into a sourcemap file.

    Example usage:

    >>> from io import StringIO
    >>> from calmjs.parse import es5
    >>> from calmjs.parse.unparsers.es5 import pretty_printer
    >>> from calmjs.parse.sourcemap import write, encode_sourcemap
    >>> program = es5(u"var i = 'hello';")
    >>> stream = StringIO()
    >>> printer = pretty_printer()
    >>> sourcemap = encode_sourcemap(
    ...     'demo.min.js', *write(printer(program), stream))
    """

    return {
        "version": 3,
        "sources": sources,
        "names": names,
        "mappings": encode_mappings(mappings),
        "file": filename,
    }


def verify_write_sourcemap_args(
        mappings, sources, names, output_stream, sourcemap_stream,
        normalize_paths=True):

    def validate_path(path, name):
        # yes, rather than equality, this token is imported from
        # the sourcemap module is the identity of all invalid
        # sources.
        if path is INVALID_SOURCE:
            # well, this was preemptively replaced, still need to
            # report this fact as a warning.
            logger.warning(
                "%s is either undefine or invalid - it is replaced "
                "with '%s'", name, INVALID_SOURCE)

    output_js = getattr(output_stream, 'name', INVALID_SOURCE)
    output_js_map = getattr(sourcemap_stream, 'name', INVALID_SOURCE)

    validate_path(output_js, 'sourcemap.file')
    validate_path(output_js_map, 'sourceMappingURL')
    for idx, source in enumerate(sources):
        validate_path(source, 'sourcemap.sources[%d]' % idx)

    if normalize_paths:
        # Caveat: macpath.pardir ignored.
        return ((
            # filename
            '/'.join(normrelpath(output_js_map, output_js).split(sep)),
            # mappings
            mappings,
            # sources
            [
                '/'.join(normrelpath(output_js_map, src).split(sep))
                for src in sources
            ],
            # names
            names,
        ), '/'.join(normrelpath(output_js, output_js_map).split(sep)))

    return (output_js, mappings, sources, names), output_js_map


def write_sourcemap(
        mappings, sources, names, output_stream, sourcemap_stream,
        normalize_paths=True, source_mapping_url=NotImplemented):
    """
    Write out the mappings, sources and names (generally produced by
    the write function) to the provided sourcemap_stream, and write the
    sourceMappingURL to the output_stream.

    Arguments

    mappings, sources, names
        These should be values produced by write function from this
        module.
    output_stream
        The original stream object that was written to; its name will
        be used for the file target and if sourceMappingURL is resolved,
        it will be writtened to this stream also as a comment.
    sourcemap_stream
        If one is provided, the sourcemap will be written out to it.

        If it is the same stream as the output_stream, the source map
        will be written as an encoded 'data:application/json;base64'
        url to the sourceMappingURL comment.  Note that an appropriate
        encoding must be available as an attribute by the output_stream
        object so that the correct character set will be used for the
        base64 encoded JSON serialized string.
    normalize_paths
        If set to True, absolute paths found will be turned into
        relative paths with relation from the stream being written
        to, and the path separator used will become a '/' (forward
        slash).
    source_mapping_url
        If an explicit value is set, this will be written as the
        sourceMappingURL into the output_stream.  Note that the path
        normalization will NOT use this value, so if paths have been
        manually provided, ensure that normalize_paths is set to False
        if the behavior is unwanted.
    """

    encode_sourcemap_args, output_js_map = verify_write_sourcemap_args(
        mappings, sources, names, output_stream, sourcemap_stream,
        normalize_paths
    )

    encoded_sourcemap = json.dumps(
        encode_sourcemap(*encode_sourcemap_args),
        sort_keys=True, ensure_ascii=False,
    )

    if sourcemap_stream is output_stream:
        # encoding will be missing if using StringIO; fall back to
        # default_encoding
        encoding = getattr(output_stream, 'encoding', None) or default_encoding
        output_stream.writelines([
            '\n//# sourceMappingURL=data:application/json;base64;charset=',
            encoding, ',', base64.b64encode(
                encoded_sourcemap.encode(encoding)).decode('ascii'),
        ])
    else:
        if source_mapping_url is not None:
            output_stream.writelines(['\n//# sourceMappingURL=', (
                output_js_map if source_mapping_url is NotImplemented
                else source_mapping_url
            ), '\n'])

        sourcemap_stream.write(encoded_sourcemap)



# -*- coding: utf-8 -*-
import unittest
from inspect import getmembers
from inspect import isfunction

from calmjs.parse import rules


class BasicRulesTestCase(unittest.TestCase):

    def test_signature_conformance(self):
        # ensure that ALL functions defined in that module are rules
        # that conform to what is expected.
        for name, f in getmembers(rules, lambda item: (
                isfunction(item) and getattr(
                    item, '__module__', None) == rules.__name__)):
            self.assertTrue(isinstance(f()(), dict))
            self.assertIn(name, rules.__all__)



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse import asttypes
from calmjs.parse.asttypes import Comments
from calmjs.parse.asttypes import BlockComment
from calmjs.parse.asttypes import LineComment
from calmjs.parse.asttypes import nodetype


class CommentNodesTestCase(unittest.TestCase):
    """
    Eventually, if a better way to have the lexer some access the
    specific node types for the syntax being processed, this would not
    be necessary.
    """

    def test_block_comment_str_repr(self):
        node = Comments([
            BlockComment(u'/* test1 */'),
            BlockComment(u'/* test2 */'),
        ])
        self.assertEqual(
            "<Comments ?children=[<BlockComment value='/* test1 */'>, "
            "<BlockComment value='/* test2 */'>]>", repr(node)
        )
        self.assertEqual("/* test1 */\n/* test2 */", str(node))

    def test_line_comment_str_repr(self):
        node = Comments([
            LineComment(u'// test1'),
            LineComment(u'// test2'),
        ])
        self.assertEqual("// test1\n// test2", str(node))


class MiscTestCase(unittest.TestCase):

    def test_nodetype(self):
        class Node(asttypes.Node):
            """
            Dummy Node
            """

        class Unsupported(asttypes.Node):
            """
            Unsupported Node
            """

        node = Node()
        unsupported = Unsupported()
        self.assertIs(asttypes.Node, nodetype(node))
        self.assertIs(Unsupported, nodetype(unsupported))
        self.assertIs(object, nodetype(object()))



# -*- coding: utf-8 -*-
"""
Provides factories to build classes.
"""

from importlib import import_module
from functools import partial
from functools import wraps
from calmjs.parse import asttypes

PKGNAME = 'calmjs.parse'  # should derive this.


class SRFactory(object):
    """
    A factory that will generate a new subclass that has the specified
    __str__ and __repr__ implementations.  Given the number of potential
    nodes that a given AST might have, tagging the custom str/repr
    on the class definition itself will only happen once, saving the
    cost of having to allocate all those references per object.
    """

    def __init__(self, module, str_, repr_):
        # recreate the class definitions
        def __str__(self):
            return str_(self)

        def __repr__(self):
            return repr_(self)

        self.module = module
        self.classes = {c.__name__: c for c in (
            type(cls.__name__, (cls,), {
                '__repr__': __repr__,
                '__str__': __str__,
            }) for cls in (
                v for v in vars(module).values() if isinstance(v, type)
            )
        )}

    def __getattr__(self, attr):
        if attr not in self.classes:
            raise AttributeError('%s "%s" has no attribute %r' % (
                type(self.module).__name__,
                self.module.__class__.__name__,
                attr,
            ))

        return self.classes[attr]


AstTypesFactory = partial(SRFactory, asttypes)


def RawParserUnparserFactory(parser_name, parse_callable, *unparse_callables):
    """
    Produces a callable object that also has callable attributes that
    passes its first argument to the parent callable.
    """

    def build_unparse(f):
        @wraps(f)
        def unparse(self, source, *a, **kw):
            node = parse_callable(
                source,
                with_comments=kw.pop('with_comments', False),
            )
            return f(node, *a, **kw)
        # a dumb and lazy docstring replacement
        if f.__doc__:
            unparse.__doc__ = f.__doc__.replace(
                'ast\n        The AST ',
                'source\n        The source ',
            )
        return unparse

    def build_parse(f):
        @wraps(f)
        def parse(self, source, *a, **kw):
            return f(source, *a, **kw)
        parse.__name__ = parser_name
        parse.__qualname__ = parser_name
        return parse

    callables = {f.__name__: build_unparse(f) for f in unparse_callables}
    callables['__call__'] = build_parse(parse_callable)
    callables['__module__'] = PKGNAME
    return type(parser_name, (object,), callables)()


def ParserUnparserFactory(module_name, *unparser_names):
    """
    Produce a new parser/unparser object from the names provided.
    """

    parse_callable = import_module(PKGNAME + '.parsers.' + module_name).parse
    unparser_module = import_module(PKGNAME + '.unparsers.' + module_name)
    return RawParserUnparserFactory(module_name, parse_callable, *[
        getattr(unparser_module, name) for name in unparser_names])



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse import vlq


class VLQTestCase(unittest.TestCase):

    def test_vlq_encode_basic(self):
        self.assertEqual(vlq.encode_vlq(0), 'A')
        self.assertEqual(vlq.encode_vlq(1), 'C')
        self.assertEqual(vlq.encode_vlq(-1), 'D')
        self.assertEqual(vlq.encode_vlq(2), 'E')
        self.assertEqual(vlq.encode_vlq(-2), 'F')

    def test_vlq_encode_edge(self):
        self.assertEqual(vlq.encode_vlq(15), 'e')
        self.assertEqual(vlq.encode_vlq(-15), 'f')
        self.assertEqual(vlq.encode_vlq(16), 'gB')
        self.assertEqual(vlq.encode_vlq(-16), 'hB')
        self.assertEqual(vlq.encode_vlq(511), '+f')
        self.assertEqual(vlq.encode_vlq(-511), '/f')
        self.assertEqual(vlq.encode_vlq(512), 'ggB')
        self.assertEqual(vlq.encode_vlq(-512), 'hgB')

    def test_vlq_encode_multi(self):
        self.assertEqual(vlq.encode_vlq(456), 'wc')
        self.assertEqual(vlq.encode_vlq(-456), 'xc')
        self.assertEqual(vlq.encode_vlq(789), 'qxB')
        self.assertEqual(vlq.encode_vlq(-789), 'rxB')

    def test_vlq_decode_various(self):
        self.assertEqual(vlq.decode_vlq('A'), 0)
        # should never be encoded to this standalone value, but nothing
        # stops this from being decoded
        self.assertEqual(vlq.decode_vlq('B'), 0)
        self.assertEqual(vlq.decode_vlq('C'), 1)
        self.assertEqual(vlq.decode_vlq('D'), -1)
        self.assertEqual(vlq.decode_vlq('E'), 2)
        self.assertEqual(vlq.decode_vlq('/f'), -511)
        self.assertEqual(vlq.decode_vlq('ggB'), 512)
        # won't decode more than one
        self.assertEqual(vlq.decode_vlq('ABCDE'), 0)

    def test_encode_vlqs(self):
        self.assertEqual(vlq.encode_vlqs((0, 1, 2, 3, 4)), 'ACEGI')
        self.assertEqual(vlq.encode_vlqs((123, 456, 789)), '2HwcqxB')

    def test_decode_vlqs(self):
        self.assertEqual((0, 1, 2, 3, 4), vlq.decode_vlqs('ACEGI'))
        self.assertEqual((123, 456, 789), vlq.decode_vlqs('2HwcqxB'))

    def test_encode_mappings(self):
        self.assertEqual(vlq.encode_mappings([
            [(0, 0, 0, 0,), (6, 0, 0, 6,), (6, 0, 0, 6,)],
            []
        ]), 'AAAA,MAAM,MAAM;')

        self.assertEqual(vlq.encode_mappings([
            [(0, 0, 0, 0,)],
            [(0, 0, 1, 0,)],
            [(0, 0, 1, 0,)],
            []
        ]), 'AAAA;AACA;AACA;')

        self.assertEqual(vlq.encode_mappings([
            [],
            [],
            [(0, 0, 0, 0,), (6, 0, 0, 6,), (6, 0, 0, 6,)],
            [(8, 0, 0, 0,)],
            [],
            [(8, 0, 2, 0,)],
            [(8, 0, 1, 0,)],
            [(8, 0, 1, 0,)],
            [],
            [],
        ]), ';;AAAA,MAAM,MAAM;QAAA;;QAEA;QACA;QACA;;')

    def test_decode_mappings(self):
        self.assertEqual([
            [(0, 0, 0, 0,), (6, 0, 0, 6,), (6, 0, 0, 6,)],
            []
        ], vlq.decode_mappings('AAAA,MAAM,MAAM;'))

        self.assertEqual([
            [(0, 0, 0, 0,)],
            [(0, 0, 1, 0,)],
            [(0, 0, 1, 0,)],
            []
        ], vlq.decode_mappings('AAAA;AACA;AACA;'))

        self.assertEqual([
            [],
            [],
            [(0, 0, 0, 0,), (6, 0, 0, 6,), (6, 0, 0, 6,)],
            [(8, 0, 0, 0,)],
            [],
            [(8, 0, 2, 0,)],
            [(8, 0, 1, 0,)],
            [(8, 0, 1, 0,)],
            [],
            [],
        ], vlq.decode_mappings(';;AAAA,MAAM,MAAM;QAAA;;QAEA;QACA;QACA;;'))



# -*- coding: utf-8 -*-
"""
Helpers for maintenance/generation of the lextab/yacctab modules.

The original goal of this was to force the creation of tab files using
the utf8 codec to workaround issues with the ply package, for systems
that do not have utf8 configured as the default codec.
"""

import codecs
import os
import sys
from functools import partial
from os import unlink
from os.path import exists
from importlib import import_module
from calmjs.parse.utils import generate_tab_names
from calmjs.parse.utils import ply_dist

_ASSUME_PLY_VERSION = '3.11'
_ASSUME_ENVVAR = 'CALMJS_PARSE_ASSUME_PLY_VERSION'


def validate_imports(*imports):
    paths = []
    missing = []
    for name in imports:
        try:
            import_module(name)
        except ImportError:
            missing.append(name)
        else:
            paths.append(sys.modules.pop(name).__file__)
    return paths, missing


def find_tab_paths(module):
    # return a list of lextab/yacctab module paths and a list of missing
    # import names.
    names = []
    for entry in ('lextab', 'yacctab'):
        # we assume the specified entries are defined as such
        name = getattr(module, entry)
        if entry not in name:
            # this really prevents this method from being a thing to
            # unlink other modules.
            raise ValueError(
                'provided module `%s` does not export expected tab values ' %
                module.__name__
            )
        names.append(name)
    return validate_imports(*names)


def purge_tabs(module):
    paths, _ = find_tab_paths(module)
    unlink_modules(verify_paths(paths))


def verify_paths(paths):
    for path in paths:
        if exists(path):
            yield path
        # locate any adjacent .py[co]? files based on module path
        # returned; mostly a problem with Python 2
        if path[-4:] in ('.pyc', '.pyo'):
            if exists(path[:-1]):
                yield path[:-1]
        else:
            for c in 'co':
                if exists(path + c):
                    yield path + c


def unlink_modules(paths):
    for path in paths:
        # will raise whatever error, i.e. when insufficient permission
        unlink(path)


def reoptimize(module):
    purge_tabs(module)
    # create a new parser should rengenerate the module
    module.Parser()


def _assume_ply_version():
    version = os.environ.get(_ASSUME_ENVVAR, _ASSUME_PLY_VERSION)
    if ply_dist is None:
        if _ASSUME_ENVVAR in os.environ:
            source = "using environment variable %r" % _ASSUME_ENVVAR
        else:
            # allow bypassing of setuptools as ply provides this
            # attribute
            try:
                import ply
                version = ply.__version__
                source = "using value provided by ply"
            except ImportError:  # pragma: no cover
                ply = None
                source = "using default value"

        sys.stderr.write(
            u"WARNING: cannot find distribution for 'ply'; "
            "%s, assuming 'ply==%s' for pre-generated modules\n" % (
                source, version))
    return version


def optimize_build(module_name, assume_ply_version=True):
    """
    optimize build helper for first build

    assume_ply_version
        This flag denotes whether or not to assume a ply version should
        ply be NOT installed; this will either assume ply to be whatever
        value assigned to _ASSUME_PLY_VERSION (i.e. 3.11), or read from
        the environment variable `CALMJS_PARSE_ASSUME_PLY_VERSION`.

        The goal is to allow the build to proceed if the pre-generated
        files are already present, before the dependency resolution at
        the installation time actually kicks in to install ply.

        Default: True
    """

    kws = {}
    if assume_ply_version:
        kws['_version'] = _assume_ply_version()

    lextab, yacctab = generate_tab_names(module_name, **kws)
    paths, missing = validate_imports(lextab, yacctab)
    if missing:
        # only import, purge and regenerate if any are missing.
        unlink_modules(verify_paths(paths))
        module = import_module(module_name)
        # use whatever assumed version or otherwise as set up by
        # the local generation function.
        module.Parser(lextab=lextab, yacctab=yacctab)


def reoptimize_all(monkey_patch=False, first_build=False):
    """
    The main optimize method for maintainence of the generated tab
    modules required by ply

    Arguments:

    monkey_patch
        patches the default open function in ply.lex to use utf8

        default: False

    first_build
        flag for switching between reoptimize/optimize_build method;
        setting the flag to True specifies the latter.

        default: False
    """

    if monkey_patch:
        try:
            from ply import lex
        except ImportError:  # pragma: no cover
            pass  # fail later; only fail if import ply is truly needed
        else:
            lex.open = partial(codecs.open, encoding='utf8')

    modules = ('.es5',)
    try:
        for name in modules:
            if first_build:
                # A consideration for modifying this flag to simply
                # check for a marker file to denote none of this being
                # needed (i.e. this tarball was fully prepared), but it
                # will not solve the issue where the distro packager
                # already got an even more recent version of ply
                # installed (as unlikely as that is) and that build step
                # then is completely skipped.
                optimize_build('calmjs.parse.parsers' + name)
            else:
                module = import_module(name, 'calmjs.parse.parsers')
                reoptimize(module)
    except ImportError as e:
        if not first_build or 'ply' not in str(e):
            raise
        sys.stderr.write(
            u"ERROR: cannot find pre-generated modules for the assumed 'ply' "
            "version from above and/or cannot `import ply` to build generated "
            "modules, aborting build; please either ensure that the source "
            "archive containing the pre-generate modules is being used, or "
            "that the python package 'ply' is installed and available for "
            "import before attempting to use the setup.py to build this "
            "package; please refer to the top level README for further "
            "details\n"
        )
        sys.exit(1)


if __name__ == '__main__':  # pragma: no cover
    reoptimize_all(True, '--build' in sys.argv)



# -*- coding: utf-8 -*-
"""
Description for ES5 unparser.
"""

from __future__ import unicode_literals
from calmjs.parse.lexers.es5 import Lexer

from calmjs.parse.ruletypes import (
    OpenBlock,
    CloseBlock,
    EndStatement,
    Space,
    RequiredSpace,
    OptionalSpace,
    Newline,
    OptionalNewline,
    Indent,
    Dedent,
    PushScope,
    PopScope,
    PushCatch,
    PopCatch,
)
from calmjs.parse.ruletypes import (
    Attr,
    CommentsAttr,
    Iter,
    Text,
    Optional,
    JoinAttr,
    Operator,
    ElisionToken,
    ElisionJoinAttr,
)
from calmjs.parse.ruletypes import (
    Literal,
    Declare,
    LineComment,
    BlockComment,
    Resolve,
    ResolveFuncName,
)
from calmjs.parse.ruletypes import (
    children_newline,
    children_comma,
)
from calmjs.parse.unparsers.base import BaseUnparser
from calmjs.parse import rules

value = (
    CommentsAttr(),
    Attr('value'),
)

# definitions of all the rules for all the types for an ES5 program.
definitions = {
    'ES5Program': (
        children_newline,
        OptionalNewline,
    ),
    'Block': (
        CommentsAttr(),
        OpenBlock,
        Indent, Newline,
        children_newline,
        Dedent, OptionalNewline,
        CloseBlock,
    ),
    'VarStatement': (
        CommentsAttr(),
        Text(value='var'), RequiredSpace, children_comma, EndStatement,
    ),
    'VarDecl': (
        CommentsAttr(),
        Attr(Declare('identifier')), Optional('initializer', (
            Space, Operator(value='='), Space, Attr('initializer'),),),
    ),
    'VarDeclNoIn': (
        CommentsAttr(),
        Text(value='var '), Attr(Declare('identifier')),
        Optional('initializer', (
            Space, Operator(value='='), Space, Attr('initializer'),),),
    ),
    'GroupingOp': (
        CommentsAttr(),
        Text(value='('), Attr('expr'), Text(value=')'),
    ),
    'Identifier': (
        CommentsAttr(),
        Attr(Resolve()),
    ),
    'PropIdentifier': value,
    'Assign': (
        CommentsAttr(),
        Attr('left'), OptionalSpace, Attr('op'), Space, Attr('right'),
    ),
    'GetPropAssign': (
        CommentsAttr(),
        Text(value='get'), RequiredSpace, Attr('prop_name'),
        PushScope,
        Text(value='('), Text(value=')'), Space,
        OpenBlock,
        Indent, Newline,
        JoinAttr(attr='elements', value=(Newline,)),
        Dedent, OptionalNewline,
        CloseBlock,
        PopScope,
    ),
    'SetPropAssign': (
        CommentsAttr(),
        Text(value='set'), RequiredSpace, Attr('prop_name'), Text(value='('),
        PushScope,
        Attr(Declare('parameter')), Text(value=')'), Space,
        OpenBlock,
        Indent, Newline,
        JoinAttr(attr='elements', value=(Newline,)),
        Dedent, OptionalNewline,
        CloseBlock,
        PopScope,
    ),
    'Number': value,
    'Comma': (
        CommentsAttr(),
        Attr('left'), Text(value=','), Space, Attr('right'),
    ),
    'EmptyStatement': (
        CommentsAttr(),
        EndStatement,
    ),
    'If': (
        CommentsAttr(),
        Text(value='if'), Space,
        Text(value='('), Attr('predicate'), Text(value=')'), Space,
        Attr('consequent'),
        Optional('alternative', (
            Newline, Text(value='else'), Space, Attr('alternative'),)),
    ),
    'Boolean': value,
    'For': (
        CommentsAttr(),
        Text(value='for'), Space, Text(value='('),
        Attr('init'),
        OptionalSpace, Attr('cond'),
        OptionalSpace, Attr('count'),
        Text(value=')'), Space, Attr('statement'),
    ),
    'ForIn': (
        CommentsAttr(),
        Text(value='for'), Space, Text(value='('),
        Attr('item'),
        RequiredSpace, Text(value='in'), Space,
        Attr('iterable'), Text(value=')'), Space, Attr('statement'),
    ),
    'BinOp': (
        CommentsAttr(),
        Attr('left'), Space, Operator(attr='op'), Space, Attr('right'),
    ),
    'UnaryExpr': (
        CommentsAttr(),
        Operator(attr='op'), OptionalSpace, Attr('value'),
    ),
    'PostfixExpr': (
        CommentsAttr(),
        Operator(attr='value'), Attr('op'),
    ),
    'ExprStatement': (
        CommentsAttr(),
        Attr('expr'), EndStatement,
    ),
    'DoWhile': (
        CommentsAttr(),
        Text(value='do'), Space, Attr('statement'), Space,
        Text(value='while'), Space, Text(value='('),
        Attr('predicate'), Text(value=')'), EndStatement,
    ),
    'While': (
        CommentsAttr(),
        Text(value='while'), Space,
        Text(value='('), Attr('predicate'), Text(value=')'), OptionalSpace,
        Attr('statement'),
    ),
    'Null': (
        CommentsAttr(),
        Text(value='null'),
    ),
    'String': (
        CommentsAttr(),
        Attr(Literal()),
    ),
    'Continue': (
        CommentsAttr(),
        Text(value='continue'), Optional('identifier', (
            RequiredSpace, Attr(attr='identifier'))), EndStatement,
    ),
    'Break': (
        CommentsAttr(),
        Text(value='break'), Optional('identifier', (
            RequiredSpace, Attr(attr='identifier'),)), EndStatement,
    ),
    'Return': (
        CommentsAttr(),
        Text(value='return'), Optional('expr', (
            Space, Attr(attr='expr'))), EndStatement,
    ),
    'With': (
        CommentsAttr(),
        Text(value='with'), Space,
        # should _really_ have a token for logging a warning
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/
        #   Reference/Statements/with
        Text(value='('), Attr('expr'), Text(value=')'), Space,
        Attr('statement'),
    ),
    'Label': (
        CommentsAttr(),
        Attr('identifier'), Text(value=':'), Space, Attr('statement'),
    ),
    'Switch': (
        CommentsAttr(),
        Text(value='switch'), Space,
        Text(value='('), Attr('expr'), Text(value=')'), Space,
        Attr('case_block'),
    ),
    'CaseBlock': (
        OpenBlock,
        Indent, Newline,
        children_newline,
        Dedent, OptionalNewline,
        CloseBlock,
    ),
    'Case': (
        CommentsAttr(),
        Text(value='case'), Space, Attr('expr'), Text(value=':'),
        Indent, Newline,
        JoinAttr('elements', value=(Newline,)),
        Dedent,
    ),
    'Default': (
        CommentsAttr(),
        Text(value='default'), Text(value=':'),
        Indent, Newline,
        JoinAttr('elements', value=(Newline,)),
        Dedent,
    ),
    'Throw': (
        CommentsAttr(),
        Text(value='throw'), Space, Attr('expr'), EndStatement,
    ),
    'Debugger': (
        CommentsAttr(),
        Text(value='debugger'), EndStatement,
    ),
    'Try': (
        CommentsAttr(),
        Text(value='try'), Space, Attr('statements'),
        Optional('catch', (Newline, Attr('catch'),)),
        Optional('fin', (Newline, Attr('fin'),)),
    ),
    'Catch': (
        CommentsAttr(),
        Text(value='catch'), Space,
        PushCatch,
        Text(value='('), Attr('identifier'), Text(value=')'), Space,
        Attr('elements'),
        PopCatch,
    ),
    'Finally': (
        CommentsAttr(),
        Text(value='finally'), Space, Attr('elements'),
    ),
    'FuncDecl': (
        CommentsAttr(),
        Text(value='function'), Optional('identifier', (RequiredSpace,)),
        Attr(Declare('identifier')), Text(value='('),
        PushScope, Optional('identifier', (ResolveFuncName,)),
        JoinAttr(Declare('parameters'), value=(Text(value=','), Space)),
        Text(value=')'), Space,
        OpenBlock,
        Indent, Newline,
        JoinAttr('elements', value=(Newline,)),
        Dedent, OptionalNewline,
        CloseBlock,
        PopScope,
    ),
    'FuncExpr': (
        CommentsAttr(),
        Text(value='function'), Optional('identifier', (RequiredSpace,)),
        Attr(Declare('identifier')), Text(value='('),
        PushScope, Optional('identifier', (ResolveFuncName,)),
        JoinAttr(Declare('parameters'), value=(Text(value=','), Space,)),
        Text(value=')'), Space,
        OpenBlock,
        Indent, Newline,
        JoinAttr('elements', value=(Newline,)),
        Dedent, OptionalNewline,
        CloseBlock,
        PopScope,
    ),
    'Conditional': (
        CommentsAttr(),
        Attr('predicate'), Space, Text(value='?'), Space,
        Attr('consequent'), Space, Text(value=':'), Space,
        Attr('alternative'),
    ),
    'Regex': value,
    'NewExpr': (
        CommentsAttr(),
        Text(value='new'), RequiredSpace, Attr('identifier'), Attr('args'),
    ),
    'DotAccessor': (
        CommentsAttr(),
        Attr('node'), Text(value='.'), Attr('identifier'),
    ),
    'BracketAccessor': (
        CommentsAttr(),
        Attr('node'), Text(value='['), Attr('expr'), Text(value=']'),
    ),
    'FunctionCall': (
        CommentsAttr(),
        Attr('identifier'), Attr('args'),
    ),
    'Arguments': (
        CommentsAttr(),
        Text(value='('),
        JoinAttr('items', value=(Text(value=','), Space)),
        Text(value=')'),
    ),
    'Object': (
        CommentsAttr(),
        Text(value='{'),
        Optional('properties', (
            Indent, Newline,
            JoinAttr('properties', value=(Text(value=','), Newline,)),
            Dedent, Newline,
        )),
        Text(value='}'),
    ),
    'Array': (
        CommentsAttr(),
        Text(value='['),
        ElisionJoinAttr('items', value=(Space,)),
        Text(value=']'),
    ),
    'Elision': (
        CommentsAttr(),
        ElisionToken(attr='value', value=','),
    ),
    'This': (
        CommentsAttr(),
        Text(value='this'),
    ),
    'Comments': (
        JoinAttr(Iter(), value=()),
    ),
    'LineComment': (
        Attr(LineComment()), Newline,
    ),
    'BlockComment': (
        Attr(BlockComment()), Newline,
    ),
}


class Unparser(BaseUnparser):

    def __init__(
            self,
            definitions=definitions,
            token_handler=None,
            rules=(rules.default(),),
            layout_handlers=None,
            deferrable_handlers=None,
            prewalk_hooks=()):

        super(Unparser, self).__init__(
            definitions=definitions,
            token_handler=token_handler,
            rules=rules,
            layout_handlers=layout_handlers,
            deferrable_handlers=deferrable_handlers,
            prewalk_hooks=prewalk_hooks,
        )


def pretty_printer(indent_str='    '):
    """
    Construct a pretty printing unparser
    """

    return Unparser(rules=(rules.indent(indent_str=indent_str),))


def pretty_print(ast, indent_str='  '):
    """
    Simple pretty print function; returns a string rendering of an input
    AST of an ES5 Program.

    arguments

    ast
        The AST to pretty print
    indent_str
        The string used for indentations.  Defaults to two spaces.
    """

    return ''.join(chunk.text for chunk in pretty_printer(indent_str)(ast))


def minify_printer(
        obfuscate=False,
        obfuscate_globals=False,
        shadow_funcname=False,
        drop_semi=False):
    """
    Construct a minimum printer.

    Arguments

    obfuscate
        If True, obfuscate identifiers nested in each scope with a
        shortened identifier name to further reduce output size.

        Defaults to False.
    obfuscate_globals
        Also do the same to identifiers nested on the global scope; do
        not enable unless the renaming of global variables in a not
        fully deterministic manner into something else is guaranteed to
        not cause problems with the generated code and other code that
        in the same environment that it will be executed in.

        Defaults to False for the reason above.
    drop_semi
        Drop semicolons whenever possible (e.g. the final semicolons of
        a given block).
    """

    active_rules = [rules.minify(drop_semi=drop_semi)]
    if obfuscate:
        active_rules.append(rules.obfuscate(
            obfuscate_globals=obfuscate_globals,
            shadow_funcname=shadow_funcname,
            reserved_keywords=(Lexer.keywords_dict.keys())
        ))
    return Unparser(rules=active_rules)


def minify_print(
        ast,
        obfuscate=False,
        obfuscate_globals=False,
        shadow_funcname=False,
        drop_semi=False):
    """
    Simple minify print function; returns a string rendering of an input
    AST of an ES5 program

    Arguments

    ast
        The AST to minify print
    obfuscate
        If True, obfuscate identifiers nested in each scope with a
        shortened identifier name to further reduce output size.

        Defaults to False.
    obfuscate_globals
        Also do the same to identifiers nested on the global scope; do
        not enable unless the renaming of global variables in a not
        fully deterministic manner into something else is guaranteed to
        not cause problems with the generated code and other code that
        in the same environment that it will be executed in.

        Defaults to False for the reason above.
    drop_semi
        Drop semicolons whenever possible (e.g. the final semicolons of
        a given block).
    """

    return ''.join(chunk.text for chunk in minify_printer(
        obfuscate, obfuscate_globals, shadow_funcname, drop_semi)(ast))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import textwrap
import unittest
from io import StringIO

from calmjs.parse import asttypes
from calmjs.parse.parsers.es5 import Parser
from calmjs.parse.parsers.es5 import parse
from calmjs.parse.parsers.es5 import read
from calmjs.parse.unparsers.es5 import pretty_print
from calmjs.parse.walkers import walk

from calmjs.parse.tests.parser import (
    ParserCaseMixin,
    build_node_repr_test_cases,
    build_asi_test_cases,
    build_syntax_error_test_cases,
    build_regex_syntax_error_test_cases,
    build_comments_test_cases,
)


class ParserTestCase(unittest.TestCase, ParserCaseMixin):

    parse = staticmethod(parse)

    def test_modify_tree(self):
        text = """
        for (var i = 0; i < 10; i++) {
          var x = 5 + i;
        }
        """
        parser = Parser()
        tree = parser.parse(text)
        for node in walk(tree):
            if isinstance(node, asttypes.Identifier) and node.value == 'i':
                node.value = 'hello'
        self.assertMultiLineEqual(
            str(tree),
            textwrap.dedent("""
            for (var hello = 0; hello < 10; hello++) {
              var x = 5 + hello;
            }
            """).lstrip()
        )

    def test_read(self):
        stream = StringIO('var foo = "bar";')
        node = read(stream)
        self.assertTrue(isinstance(node, asttypes.ES5Program))
        self.assertIsNone(node.sourcepath)

        stream.name = 'somefile.js'
        node = read(stream)
        self.assertEqual(node.sourcepath, 'somefile.js')


ParsedNodeTypeTestCase = build_node_repr_test_cases(
    'ParsedNodeTypeTestCase', parse, 'ES5Program')

# ASI - Automatic Semicolon Insertion
ParserToECMAASITestCase = build_asi_test_cases(
    'ParserToECMAASITestCase', parse, pretty_print)

ECMASyntaxErrorsTestCase = build_syntax_error_test_cases(
    'ECMASyntaxErrorsTestCase', parse)

ECMARegexSyntaxErrorsTestCase = build_regex_syntax_error_test_cases(
    'ECMARegexSyntaxErrorsTestCase', parse)

ParsedNodeTypesWithCommentsTestCase = build_comments_test_cases(
    'ParsedNodeTypeWithCommentsTestCase', parse, 'ES5Program')



###############################################################################
# encoding: utf-8
#
# Copyright (c) 2011 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

import unittest
import textwrap
from functools import partial

from calmjs.parse.lexers.es5 import Lexer
from calmjs.parse.exceptions import ECMASyntaxError

from calmjs.parse.testing.util import build_equality_testcase
from calmjs.parse.testing.util import build_exception_testcase
from calmjs.parse.tests.lexer import (
    run_lexer,
    run_lexer_pos,
    es5_cases,
    es5_all_cases,
    es5_pos_cases,
    es5_error_cases_str_sq,
    es5_error_cases_str_dq,
)


class LexerFailureTestCase(unittest.TestCase):

    def test_illegal_unicode_char_in_identifier(self):
        lexer = Lexer()
        lexer.input(u'\u0036_tail')
        token = lexer.token()
        # \u0036_tail is the same as 6_tail and that's not a correct ID
        # Check that the token is NUMBER and not an ID
        self.assertEqual(token.type, 'NUMBER')
        self.assertEqual(token.value, '6')

    def test_bad_initial_input(self):
        lexer = Lexer()
        lexer.input(u'#')
        with self.assertRaises(ECMASyntaxError) as e:
            lexer.token()

        self.assertEqual(e.exception.args[0], "Illegal character '#' at 1:1")

    def test_extra_ending_braces(self):
        lexer = Lexer()
        lexer.input('\n\n())))')
        with self.assertRaises(ECMASyntaxError) as e:
            [token for token in lexer]
        self.assertEqual(str(e.exception), "Mismatched ')' at 3:3")


class LexerGeneralTestCase(unittest.TestCase):

    def test_backtracking_div_default(self):
        # this test emulates what appear to happens when the parser
        # backtracks upon encountering the `/` that it can't parse.
        lexer = Lexer()
        lexer.input('{}/a/')
        self.assertEqual(lexer.next().type, 'LBRACE')
        self.assertEqual(lexer.next().type, 'RBRACE')
        self.assertEqual(lexer.next().type, 'DIV')

        token = lexer.backtracked_token()
        self.assertEqual(('REGEX', '/a/'), (token.type, token.value))

    def test_backtracking_multiple(self):
        # Although dealing with additional tokens like comments and
        # newlines are not done (i.e. they don't get backtracked), it
        # should not cause additional issues
        lexer = Lexer()
        lexer.input('{}\n/a/')
        self.assertEqual(lexer.next().type, 'LBRACE')
        self.assertEqual(lexer.next().type, 'RBRACE')
        self.assertEqual(lexer.next().type, 'DIV')

        token = lexer.backtracked_token(pos=2)
        self.assertEqual(('REGEX', '/a/'), (token.type, token.value))


class LexerWithCommentsTestCase(unittest.TestCase):

    def test_with_line_comments_before(self):
        lexer = Lexer(with_comments=True)
        lexer.input(textwrap.dedent('''
        // foo
        // bar
        baz
        '''))
        tokens = [token for token in lexer]
        self.assertEqual(1, len(tokens))
        token = tokens[0]
        self.assertEqual(2, len(token.hidden_tokens))
        self.assertEqual(token.hidden_tokens[0].value, '// foo')
        self.assertEqual(token.hidden_tokens[0].type, 'LINE_COMMENT')
        self.assertEqual(token.hidden_tokens[1].value, '// bar')
        self.assertEqual(token.hidden_tokens[1].type, 'LINE_COMMENT')

    def test_with_line_comment_trail(self):
        lexer = Lexer(with_comments=True)
        lexer.input(textwrap.dedent('''
        baz // bar
        foo
        '''))
        tokens = [token for token in lexer]
        self.assertEqual(2, len(tokens))
        self.assertFalse(hasattr(tokens[0], 'hidden_tokens'))
        token = tokens[1]
        self.assertEqual(1, len(token.hidden_tokens))
        self.assertEqual(token.hidden_tokens[0].value, '// bar')
        self.assertEqual(token.hidden_tokens[0].type, 'LINE_COMMENT')

    def test_block_comment(self):
        lexer = Lexer(with_comments=True)
        lexer.input(textwrap.dedent('''
        baz /*foo
        *// // bar
        '''))
        tokens = [token for token in lexer]
        self.assertEqual(2, len(tokens))
        self.assertFalse(hasattr(tokens[0], 'hidden_tokens'))
        token = tokens[1]
        self.assertEqual(1, len(token.hidden_tokens))
        self.assertEqual(token.hidden_tokens[0].value, '/*foo\n*/')
        self.assertEqual(token.hidden_tokens[0].type, 'BLOCK_COMMENT')
        # the remaining comments remain on lexer
        self.assertEqual(lexer.hidden_tokens[0].value, '// bar')


LexerKeywordTestCase = build_equality_testcase(
    'LexerKeywordTestCase', partial(run_lexer, lexer_cls=Lexer), (
        (label, data[0], data[1],) for label, data in [(
            # Keywords
            # ('break case ...', ['BREAK break', 'CASE case', ...])
            'keywords_all',
            (' '.join(kw.lower() for kw in Lexer.keywords),
             ['%s %s' % (kw, kw.lower()) for kw in Lexer.keywords]
             ),
        )]
    )
)

LexerTestCase = build_equality_testcase(
    'LexerTestCase', partial(run_lexer, lexer_cls=Lexer), (
        (label, data[0], data[1],) for label, data in es5_cases))

LexerAllTestCase = build_equality_testcase(
    'LexerAllTestCase', partial(run_lexer, lexer_cls=partial(
        Lexer, yield_comments=True
    )), ((label, data[0], data[1],) for label, data in es5_all_cases))

LexerErrorTestCase = build_exception_testcase(
    'LexerErrorTestCase', partial(
        run_lexer, lexer_cls=Lexer), es5_error_cases_str_sq, ECMASyntaxError)

LexerErrorStrDQTestCase = build_exception_testcase(
    'LexerErrorStrDQTestCase', partial(
        run_lexer, lexer_cls=Lexer), es5_error_cases_str_dq, ECMASyntaxError)

LexerPosTestCase = build_equality_testcase(
    'LexerPosTestCase', partial(
        run_lexer_pos, lexer_cls=Lexer), es5_pos_cases)



# -*- coding: utf-8 -*-
"""
Various utilities and helpers
"""

import sys
from os.path import dirname
from os.path import isabs
from os.path import normpath
from os.path import relpath


class _Distribution(object):
    def __init__(self, project_name, version):
        self.project_name = project_name
        self.version = version


try:
    from importlib import metadata
    try:
        ply_version = metadata.version('ply')
    except Exception:  # pragma: no cover
        ply_dist = None
    else:
        ply_dist = _Distribution(project_name='ply', version=ply_version)
except ImportError:  # pragma: no cover
    try:
        from pkg_resources import working_set
        from pkg_resources import Requirement
        ply_dist = working_set.find(Requirement.parse('ply'))
        # note that for **extremely** ancient versions of setuptools, e.g.
        # setuptools<0.6c11, or some very non-standard environment that does
        # not include the required metadata (e.g. pyinstaller without the
        # required metadata), will require the following workaround...
        if ply_dist:
            # convert to our private version class
            ply_dist = _Distribution(
                project_name='ply',
                version=ply_dist.version,
            )
        else:
            try:
                import ply
                ply_dist = _Distribution(
                    project_name='ply',
                    version=ply.__version__,
                )
            except ImportError:
                ply_dist = None
    except ImportError:  # pragma: no cover
        ply_dist = None

py_major = sys.version_info.major
unicode = unicode if py_major < 3 else None  # noqa: F821
str = str if sys.version_info.major > 2 else unicode  # noqa: F821


def repr_compat(s):
    """
    Since Python 2 is annoying with unicode literals, and that we are
    enforcing the usage of unicode, this ensures the repr doesn't spew
    out the unicode literal prefix.
    """

    if unicode and isinstance(s, unicode):
        return repr(s)[1:]
    else:
        return repr(s)


def generate_tab_names(name, _version='unknown'):
    """
    Return the names to lextab and yacctab modules for the given module
    name.  Typical usage should be like so::

    >>> lextab, yacctab = generate_tab_names(__name__)
    """

    package_name, module_name = name.rsplit('.', 1)

    version = (ply_dist.version if ply_dist is not None else _version).replace(
        '.', '_')
    data = (package_name, module_name, py_major, version)
    lextab = '%s.lextab_%s_py%d_ply%s' % data
    yacctab = '%s.yacctab_%s_py%d_ply%s' % data
    return lextab, yacctab


def format_lex_token(token):
    return '%s at %s:%s' % (
        repr_compat(token.value), token.lineno, getattr(token, 'colno', '?'))


def normrelpath(base, target):
    """
    This function takes the base and target arguments as paths, and
    returns an equivalent relative path from base to the target, if both
    provided paths are absolute.
    """

    if not all(map(isabs, [base, target])):
        return target

    return relpath(normpath(target), dirname(normpath(base)))



# -*- coding: utf-8 -*-
"""
Rules setup functions for standard Unparser constructors.  All functions
here return a rule setup function that returns a working and complete
set of handlers for the desired functionality as designed.  While the
rules defined here can mix, note that the Unparser will treat the later
rules as having prescedence for any rule conflicts.
"""

from calmjs.parse.ruletypes import (
    OpenBlock,
    CloseBlock,
    EndStatement,
    Space,
    OptionalSpace,
    RequiredSpace,
    Newline,
    OptionalNewline,
    Indent,
    Dedent,
    Resolve,
    Literal,

    LineComment,
    BlockComment,
)
from calmjs.parse.handlers.core import (
    rule_handler_noop,

    token_handler_unobfuscate,

    layout_handler_openbrace,
    layout_handler_closebrace,
    layout_handler_semicolon,
    layout_handler_semicolon_optional,

    layout_handler_space_imply,
    layout_handler_space_optional_pretty,
    layout_handler_space_minimum,

    deferrable_handler_literal_continuation,
    deferrable_handler_comment,

    default_rules,
    minimum_rules,
)
from calmjs.parse.handlers.indentation import Indentator
from calmjs.parse.handlers.obfuscation import Obfuscator

__all__ = ['default', 'minimum', 'minify', 'indent', 'obfuscate']


def default():
    """
    A default set of handlers
    """

    return default_rules


def minimum():
    """
    The minimum required set of handlers for the results to still be
    considered valid code.  This is due to the fact that a set of
    minimum whitespace handlers are still needed.
    """

    return minimum_rules


def minify(drop_semi=True):
    """
    Rules for minifying output.

    Arguments:

    drop_semi
        Drop semicolons whenever possible.  Note that if Dedent and
        OptionalNewline has a handler defined, it will stop final break
        statements from being resolved due to reliance on normalized
        resolution.

    """

    layout_handlers = {
        OpenBlock: layout_handler_openbrace,
        CloseBlock: layout_handler_closebrace,
        EndStatement: layout_handler_semicolon,
        Space: layout_handler_space_minimum,
        OptionalSpace: layout_handler_space_minimum,
        RequiredSpace: layout_handler_space_imply,
        (Space, OpenBlock): layout_handler_openbrace,
        (Space, EndStatement): layout_handler_semicolon,
        (OptionalSpace, EndStatement): layout_handler_semicolon,
    }

    if drop_semi:
        # if these are defined, they should be dropped; should really
        # provide these as a flag.
        # layout_handlers.update({
        #     OptionalNewline: None,
        #     Dedent: None,
        # })

        layout_handlers.update({
            EndStatement: layout_handler_semicolon_optional,

            # these two rules rely on the normalized resolution
            (OptionalSpace, EndStatement): layout_handler_semicolon_optional,
            (EndStatement, CloseBlock): layout_handler_closebrace,

            # this is a fallback rule for when Dedent is defined by
            # some other rule, which won't neuter all optional
            # semicolons.
            (EndStatement, Dedent): rule_handler_noop,
            ((OptionalSpace, EndStatement), CloseBlock):
                layout_handler_closebrace,
        })

    def minify_rule():
        return {
            'layout_handlers': layout_handlers,
            'deferrable_handlers': {
                Literal: deferrable_handler_literal_continuation,
            },
        }

    return minify_rule


def indent(indent_str=None):
    """
    A complete, standalone indent ruleset.

    Arguments:

    indent_str
        The string used for indentation.  Defaults to None, which will
        defer the value used to the one provided by the Dispatcher.
    """

    def indentation_rule():
        inst = Indentator(indent_str)
        return {'layout_handlers': {
            OpenBlock: layout_handler_openbrace,
            CloseBlock: layout_handler_closebrace,
            EndStatement: layout_handler_semicolon,
            Space: layout_handler_space_imply,
            OptionalSpace: layout_handler_space_optional_pretty,
            RequiredSpace: layout_handler_space_imply,
            Indent: inst.layout_handler_indent,
            Dedent: inst.layout_handler_dedent,
            Newline: inst.layout_handler_newline,
            OptionalNewline: inst.layout_handler_newline_optional,
            (Space, OpenBlock): NotImplemented,
            (Space, EndStatement): layout_handler_semicolon,
            (OptionalSpace, EndStatement): layout_handler_semicolon,
            (Indent, Newline, Dedent): rule_handler_noop,
        }, 'deferrable_handlers': {
            LineComment: deferrable_handler_comment,
            BlockComment: deferrable_handler_comment,
        }}
    return indentation_rule


def obfuscate(
        obfuscate_globals=False,
        shadow_funcname=False,
        reserved_keywords=()):
    """
    The name obfuscation ruleset.

    Arguments:

    obfuscate_globals
        If true, identifier names on the global scope will also be
        obfuscated.  Default is False.
    shadow_funcname
        If True, permit the shadowing of the name of named functions by
        names within the scope it defines.  Default is False.
    reserved_keywords
        A tuple of strings that should not be generated as obfuscated
        identifiers.
    """

    def name_obfuscation_rules():
        inst = Obfuscator(
            obfuscate_globals=obfuscate_globals,
            shadow_funcname=shadow_funcname,
            reserved_keywords=reserved_keywords,
        )
        return {
            'token_handler': token_handler_unobfuscate,
            'layout_handlers': {
            },
            'deferrable_handlers': {
                Resolve: inst.resolve,
            },
            'prewalk_hooks': [
                inst.prewalk_hook,
            ],
        }
    return name_obfuscation_rules



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse.asttypes import Node
from calmjs.parse.handlers.core import (
    layout_handler_space_optional_pretty,
    layout_handler_space_minimum,
    layout_handler_newline_optional_pretty,
    layout_handler_openbrace,
    layout_handler_closebrace,
    layout_handler_semicolon,
    deferrable_handler_literal_continuation,
)
from calmjs.parse.unparsers.walker import Dispatcher

empty = []
space = [(' ', 0, 0, None, None)]


class SimpleHandlersTestCase(unittest.TestCase):
    """
    Assorted test cases for edges.
    """

    def test_core_structures(self):
        # initialise a barebone dispatcher.
        node = Node()
        dispatcher = Dispatcher({}, None, {}, {})
        self.assertEqual([('{', 0, 0, None, None,)], list(
            layout_handler_openbrace(dispatcher, node, None, None, None)))
        self.assertEqual([('}', 0, 0, None, None,)], list(
            layout_handler_closebrace(dispatcher, node, None, None, None)))
        self.assertEqual([(';', 0, 0, None, None,)], list(
            layout_handler_semicolon(dispatcher, node, None, None, None)))

        # with token map
        node._token_map = {
            '{': [(0, 1, 1)],
            '}': [(1, 1, 2)],
            ';': [(2, 1, 3)],
        }
        self.assertEqual([('{', 1, 1, None, None,)], list(
            layout_handler_openbrace(dispatcher, node, None, None, None)))
        self.assertEqual([('}', 1, 2, None, None,)], list(
            layout_handler_closebrace(dispatcher, node, None, None, None)))
        self.assertEqual([(';', 1, 3, None, None,)], list(
            layout_handler_semicolon(dispatcher, node, None, None, None)))

    def test_space_optional_pretty(self):
        # initialise a barebone dispatcher.
        dispatcher = Dispatcher({}, None, {}, {})

        def run(a, b):
            return list(layout_handler_space_optional_pretty(
                # node and prev are not used.
                dispatcher, None, a, b, None))

        # also test out the cases where OptionalSpace was defined.
        self.assertEqual(run(None, None), empty)
        self.assertEqual(run('a', None), empty)
        self.assertEqual(run(None, 'a'), empty)
        # for Assign
        self.assertEqual(run('a', ':'), empty)
        self.assertEqual(run('a', '='), space)
        self.assertEqual(run('1', '='), space)

        self.assertEqual(run('a', '+='), space)
        self.assertEqual(run('a', '-='), space)
        self.assertEqual(run('a', '*='), space)
        self.assertEqual(run('a', '/='), space)
        self.assertEqual(run('a', '%='), space)
        self.assertEqual(run('a', '&='), space)
        self.assertEqual(run('a', '^='), space)
        self.assertEqual(run('a', '|='), space)
        self.assertEqual(run('a', '<<='), space)
        self.assertEqual(run('a', '>>='), space)
        self.assertEqual(run('a', '>>>='), space)

        # these rules are not defined, since they typically shouldn't
        # happen and that BinOp rules should use Space.
        self.assertEqual(run('+', 'a'), empty)
        self.assertEqual(run('-', 'a'), empty)
        self.assertEqual(run('*', 'a'), empty)
        self.assertEqual(run('/', 'a'), empty)

        # for Unary
        self.assertEqual(run('!', 'a'), empty)
        self.assertEqual(run('f', '1'), space)
        self.assertEqual(run('1', 'f'), space)
        self.assertEqual(run('f', '++'), empty)
        self.assertEqual(run('f', '--'), empty)
        self.assertEqual(run('--', 'f'), empty)
        self.assertEqual(run('++', 'f'), empty)
        self.assertEqual(run('+', '-'), empty)
        self.assertEqual(run('-', '+'), empty)
        self.assertEqual(run('-', '-'), space)
        self.assertEqual(run('+', '+'), space)

    def test_space_minimum(self):
        # initialise a barebone dispatcher.
        dispatcher = Dispatcher({}, None, {}, {})

        def run(a, b):
            return list(layout_handler_space_minimum(
                # node and prev are not used.
                dispatcher, None, a, b, None))

        self.assertEqual(run(None, None), empty)

        # for Assign
        self.assertEqual(run('a', ':'), empty)
        self.assertEqual(run('a', '='), empty)

        # for Unary
        self.assertEqual(run('!', 'a'), empty)
        self.assertEqual(run('f', '1'), space)
        self.assertEqual(run('1', 'f'), space)
        self.assertEqual(run('f', '+'), empty)
        self.assertEqual(run('f', '-'), empty)
        self.assertEqual(run('-', 'f'), empty)
        self.assertEqual(run('+', 'f'), empty)
        self.assertEqual(run('+', '-'), empty)
        self.assertEqual(run('-', '-'), space)
        self.assertEqual(run('+', '+'), space)
        self.assertEqual(run(',', 'y'), empty)
        self.assertEqual(run('1', ','), empty)

    def test_layout_handler_newline_optional_pretty(self):
        # yes using the lteral <CR><LF> is pretty hilarious, but just to
        # show that this is implemented to support whatever.
        dispatcher = Dispatcher({}, None, {}, {}, newline_str='<CR><LF>')
        newline = [('<CR><LF>', 0, 0, None, None)]

        def run(before, after, prev):
            return list(layout_handler_newline_optional_pretty(
                # node is not used.
                dispatcher, None, before, after, prev))

        # brand new empty program
        self.assertEqual(run(None, None, None), empty)
        # first line of a program
        self.assertEqual(run(None, 'function', None), empty)
        # previous token produced a newline
        self.assertEqual(run('\n', 'function', None), empty)
        # next token produced a newline
        self.assertEqual(run('}', '\n', None), empty)
        # Previous layout produced a newline
        self.assertEqual(run('}', '  ', '\n'), empty)

        # Previous layout produced a newline, with the custom str
        self.assertEqual(run('}', '  ', '<CR><LF>'), empty)

        # The first layout rule
        self.assertEqual(run(';', 'function', None), newline)

    def test_deferrable_handler_literal_continuation(self):
        dispatcher = Dispatcher({}, None, {}, {})
        node = Node()
        node.value = '"foo\\\r\nbar"'
        self.assertEqual('"foobar"', deferrable_handler_literal_continuation(
            dispatcher, node))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
import unittest
import doctest
from textwrap import dedent
from io import StringIO
from os.path import basename
from os.path import dirname

examples = {
    '/tmp/html4.js': dedent("""
    var bold = function(s) {
        return '<b>' + s + '</b>';
    };

    var italics = function(s) {
        return '<i>' + s + '</i>';
    };
    """).lstrip(),
    '/tmp/legacy.js': dedent("""
    var marquee = function(s) {
        return '<marquee>' + s + '</marquee>';
    };

    var blink = function(s) {
        return '<blink>' + s + '</blink>';
    };
    """).lstrip(),
}


class StringIOWrapper(StringIO):
    """
    Needed to rstrip the output to make all doctest output behave in a
    consistent way across all platforms because for whatever reason the
    doctest has different behaviors between Windows and others...
    """

    def read(self):
        return StringIO.read(self).rstrip()


def make_suite():  # pragma: no cover
    from calmjs.parse.lexers import es5 as es5lexer
    from calmjs.parse import walkers
    from calmjs.parse import sourcemap

    def open(p, flag='r'):
        result = StringIOWrapper(examples[p] if flag == 'r' else '')
        # Need basename here because Python 3.13 under Windows broke
        # _something_ and made the reporting inconsistent...
        result.name = basename(p)
        return result

    parser = doctest.DocTestParser()
    # Python 2 can't handle current string standards.
    optflags = (
        doctest.SKIP
        if sys.version_info < (3,) else
        doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    )

    try:
        from importlib import metadata
        pkgdesc = metadata.metadata('calmjs.parse').get('description')
        pkgdesc = pkgdesc.replace('\r', '') if pkgdesc else ''
    except ImportError:
        from pkg_resources import get_distribution
        dist = get_distribution('calmjs.parse')
        if dist:
            if dist.has_metadata('PKG-INFO'):
                pkgdesc = dist.get_metadata('PKG-INFO').replace('\r', '')
            elif dist.has_metadata('METADATA'):
                pkgdesc = dist.get_metadata('METADATA').replace('\r', '')
            else:
                pkgdesc = ''

    pkgdesc_tests = [
        t for t in parser.parse(pkgdesc) if isinstance(t, doctest.Example)]

    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(
        'calmjs.parse.tests', pattern='test_*.py',
        top_level_dir=dirname(__file__)
    )
    try:
        test_suite.addTest(doctest.DocTestSuite(
            es5lexer, optionflags=optflags))
        test_suite.addTest(doctest.DocTestSuite(
            walkers, optionflags=optflags))
        test_suite.addTest(doctest.DocTestSuite(
            sourcemap, optionflags=optflags))
        test_suite.addTest(doctest.DocTestCase(
            # skipping all the error case tests which should all be in the
            # troubleshooting section at the end; bump the index whenever
            # more failure examples are added.
            # also note that line number is unknown, as PKG_INFO has headers
            # and also the counter is somehow inaccurate in this case.
            doctest.DocTest(pkgdesc_tests[:-1], {
                'open': open}, 'PKG_INFO', 'README.rst', None, pkgdesc),
            optionflags=optflags,
        ))
    except AttributeError:
        # Assuming this is in Python>3.9 where the -OO flag was used...
        pass

    return test_suite



# -*- coding: utf-8 -*-
"""
Quick access helper functions
"""

try:
    from calmjs.parse.factory import ParserUnparserFactory
except ImportError as e:  # pragma: no cover
    exc = e

    def import_error(*a, **kw):
        raise exc

    es5 = import_error
else:
    es5 = ParserUnparserFactory('es5', 'pretty_print', 'minify_print')
